#
# Copyright 2021 Johannes Hörmann
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import datetime
import logging
import os

from gi.repository import Gdk, GLib, Gio, Gtk, GtkSource

from ..utils.logging import FormattedPrependingGtkTextBufferHandler

logger = logging.getLogger(__name__)


@Gtk.Template(filename=f'{os.path.dirname(__file__)}/log_window.ui')
class LogWindow(Gtk.Window):
    __gtype_name__ = 'DtoolLogWindow'

    log_text_view = Gtk.Template.Child()

    clear_button = Gtk.Template.Child()
    copy_button = Gtk.Template.Child()
    save_button = Gtk.Template.Child()

    log_switch = Gtk.Template.Child()
    # loglevel_entry = Gtk.Template.Child()
    loglevel_combo_box = Gtk.Template.Child()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        root_logger = logging.getLogger()

        self.clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)

        # populate logleve selection combo box
        loglevel_store = Gtk.ListStore(int, str)
        configurable_loglevels = [
            (logging.CRITICAL, "CRITICAL"),
            (logging.ERROR, "ERROR"),
            (logging.WARNING, "WARNING"),
            (logging.INFO, "INFO"),
            (logging.DEBUG, "DEBUG"),
        ]

        self.loglevel_row_index_map = {}
        for row_index, (loglevel_value, loglevel_label) in enumerate(configurable_loglevels):
            loglevel_store.append([loglevel_value, loglevel_label])
            self.loglevel_row_index_map[loglevel_value] = row_index

        self.loglevel_combo_box.set_model(loglevel_store)
        self.loglevel_combo_box.set_active(self.loglevel_row_index_map[root_logger.level])
        self.loglevel_combo_box.connect("changed", self.on_loglevel_combo_box_changed)
        self.loglevel_combo_box.set_entry_text_column(1)

        # connect log handler to my buffer
        self.log_buffer = self.log_text_view.get_buffer()
        self.log_handler = FormattedPrependingGtkTextBufferHandler(
            text_buffer=self.log_buffer)

        # set up some pseudo highlighting for the log window
        lang_manager = GtkSource.LanguageManager()
        self.log_buffer.set_language(lang_manager.get_language("python"))
        self.log_buffer.set_highlight_syntax(True)
        self.log_buffer.set_highlight_matching_brackets(True)

        logger.debug("Append GtkTextBufferHandler to root logger.")
        root_logger.addHandler(self.log_handler)

        # bind another handler to th application-level action 'set-loglevel'
        set_loglevel_action = self.get_action_group("app").lookup_action('set-loglevel')
        set_loglevel_action.connect("change-state", self.do_loglevel_changed)

        root_logger.debug("Created log window.")

    # action handlers
    def do_loglevel_changed(self, action, value):
        """Keep entry ot loglevel combobox in sync with actual action state."""
        new_loglevel = value.get_uint16()
        logger.debug(f"Loglevel changed to {new_loglevel}, update combo box entry if necessary.")
        tree_iter = self.loglevel_combo_box.get_active_iter()
        model = self.loglevel_combo_box.get_model()
        current_loglevel, loglevel_label = model[tree_iter][:2]
        logger.debug(f"Current loglevel is {current_loglevel}.")
        if new_loglevel != current_loglevel:
            self.loglevel_combo_box.set_active(self.loglevel_row_index_map[new_loglevel])
            logger.debug(f"Combo box updated to row {self.loglevel_row_index_map[new_loglevel]}.")

    # signal handlers
    @Gtk.Template.Callback()
    def on_show(self, widget):
        logger.debug("Show.")
        root_logger = logging.getLogger()

    @Gtk.Template.Callback()
    def on_delete(self, widget, event):
        logger.debug("Hide on delete.")
        return self.hide_on_delete()

    @Gtk.Template.Callback()
    def on_destroy(self, widget):
        logger.debug("Destroy.")
        root_logger = logging.getLogger()
        if self.log_handler in root_logger.handlers:
            root_logger.removeHandler(self.log_handler)

    @Gtk.Template.Callback()
    def on_log_switch_state_set(self, widget, state):
        logger.debug(f"{widget.get_name()} switched state to {state}")
        # Eventually managed to tie switch directly to action via glade
        # by setting up a stateful but parameterless toggle action
        # and specifying app.toggle-logging as action name for switch in xml,
        # hence no need for this handler anymore.

    @Gtk.Template.Callback()
    def on_loglevel_combo_box_changed(self, combo):
        """Selected loglevel from combo box."""
        tree_iter = combo.get_active_iter()
        model = combo.get_model()
        loglevel, loglevel_label = model[tree_iter][:2]
        logger.debug(f"Selected ID: {loglevel}, loglevel: {loglevel_label}")
        # This explicitly evokes the according action when loglevel selected
        # in combo box turned, see
        # https://lazka.github.io/pgi-docs/Gio-2.0/classes/ActionGroup.html#Gio.ActionGroup.list_actions
        # There might be more elegant mechanism to connect a switch with an
        # app-central action, but the Gtk docs are sparse on actions...
        self.get_action_group("app").activate_action('set-loglevel', GLib.Variant.new_uint16(loglevel))

    @Gtk.Template.Callback()
    def on_clear_clicked(self, widget):
        """Clear contents of log window."""
        self.log_buffer.set_text("")

    @Gtk.Template.Callback()
    def on_copy_clicked(self, widget):
        """Copy contents of log window to clipboard."""
        start_iter = self.log_buffer.get_start_iter()
        end_iter = self.log_buffer.get_end_iter()
        self.log_buffer.select_range(start_iter, end_iter)
        self.log_buffer.copy_clipboard(self.clipboard)
        logger.debug("Copied content of log window to clipboard.")

    @Gtk.Template.Callback()
    def on_save_clicked(self, widget):
        """Open file chooser dialog and save contents of log window to file."""
        dialog = Gtk.FileChooserDialog(title=f"Save log", parent=self,
                                       action=Gtk.FileChooserAction.SAVE)
        dialog.add_buttons(Gtk.STOCK_CANCEL,
                           Gtk.ResponseType.CANCEL,
                           Gtk.STOCK_OK,
                           Gtk.ResponseType.OK)
        suggested_file_name = f"{datetime.datetime.now().isoformat()}-{self.get_application().get_application_id()}.log"
        dialog.set_current_name(suggested_file_name)
        dialog.set_do_overwrite_confirmation(True)

        response = dialog.run()

        if response == Gtk.ResponseType.OK:
            start_iter = self.log_buffer.get_start_iter()
            end_iter = self.log_buffer.get_end_iter()
            dest_filename = dialog.get_filename()
            with open(dest_filename, "w") as file:
                file.write(self.log_buffer.get_text(start_iter, end_iter, include_hidden_chars=False))

        elif response == Gtk.ResponseType.CANCEL:
            pass
        dialog.destroy()



#
# Copyright 2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from gi.repository import Gtk

from dtool_info.utils import date_fmt


class DtoolDatasetRow(Gtk.ListBoxRow):
    __gtype_name__ = 'DtoolDatasetRow'

    _margin = 3

    def __init__(self, dataset, *args, **kwargs):
        self._dataset = dataset

        super().__init__(*args, **kwargs)

        self._build()

    def _build(self):
        # Clear current contents
        for child in self.get_children():
            child.destroy()

        # Update row contents
        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, margin_top=self._margin, margin_bottom=self._margin,
                       margin_start=self._margin, margin_end=self._margin)
        self.uuid_label = Gtk.Label(xalign=0)
        vbox.pack_start(self.uuid_label, True, True, 0)
        self.name_label = Gtk.Label(xalign=0)
        vbox.pack_start(self.name_label, True, True, 0)
        self.info_label = Gtk.Label(xalign=0)
        vbox.pack_start(self.info_label, True, True, 0)
        self.add(vbox)

        self._refresh()

    def _refresh(self):
        if self._dataset.is_frozen:
            self.uuid_label.set_markup(f'<b>{self._dataset.uuid}</b>')
        else:
            self.uuid_label.set_markup(f'<b>* {self._dataset.uuid}</b>')
        self.name_label.set_markup(f'{self._dataset.name}')
        self.info_label.set_markup(
            f'<small>Created by {self._dataset.creator}, ' +
            (f'frozen at {self._dataset.date}, {self._dataset.size_str.strip()}'
             if self._dataset.is_frozen else 'not yet frozen') +
            f'</small>')

    @property
    def dataset(self):
        return self._dataset

    def freeze(self):
        self._dataset.freeze()
        self._refresh()


#
# Copyright 2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import locale
from contextlib import contextmanager
from datetime import date, datetime

import dtoolcore.utils


@contextmanager
def time_locale(name):
    # This code snippet was taken from:
    # https://stackoverflow.com/questions/18593661/how-do-i-strftime-a-date-object-in-a-different-locale
    saved = locale.setlocale(locale.LC_TIME)
    try:
        yield locale.setlocale(locale.LC_TIME, name)
    finally:
        locale.setlocale(locale.LC_TIME, saved)


def to_timestamp(d):
    """
    Convert a string or a timestamp to a timestamp. This is a dirty fix necessary
    because the /dataset/list route return timestamps but /dataset/search
    returns strings in older versions of the lookup server (before 0.15.0).
    """
    if type(d) is str:
        try:
            with time_locale('C'):
                d = dtoolcore.utils.timestamp(datetime.strptime(d, '%a, %d %b %Y %H:%M:%S %Z'))
        except ValueError as e:
            d = -1
    return d


def date_to_string(d):
    return date.fromtimestamp(to_timestamp(d))



#
# Copyright 2021 Johannes Hörmann
#           2020-2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import logging

from math import pi, sqrt
_sqrt3 = sqrt(3)

import numpy as np

from gi.repository import GLib, GObject, Gdk, Gtk

from ..models.simple_graph import GraphLayout
from ..utils.query import dump_single_line_query_text

from .graph_popover import DtoolGraphPopover


logger = logging.getLogger(__name__)

# with a timeout of 10 microseconds, GUI freezes regularly on my machine (Ubuntu 20.04)
TIMEOUT = 50


def circle(context, x, y):
    context.arc(x, y, 0.5, 0, 2 * pi)
    context.close_path()


def square(context, x, y):
    context.rectangle(x - 0.4, y - 0.4, 0.8, 0.8)


def triangle(context, x, y):
    height = 0.5*_sqrt3
    context.move_to(x, y - 0.5*height)
    context.line_to(x + 0.5, y + 0.5*height)
    context.line_to(x - 0.5, y + 0.5*height)
    context.close_path()


class DtoolGraphWidget(Gtk.DrawingArea):
    __gtype_name__ = 'DtoolGraphWidget'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._timer = None
        self._graph = None
        self._layout = None

        self._search_by_uuid = None

        self._popover = DtoolGraphPopover(on_show_clicked=self.on_show_clicked)
        self._popover.set_relative_to(self)

        self.connect('realize', self.on_realize)
        self.connect('draw', self.on_draw)
        self.connect('motion-notify-event', self.on_motion_notify)
        self.set_events(Gdk.EventMask.POINTER_MOTION_MASK)

    @property
    def search_by_uuid(self):
        return self._search_by_uuid

    @search_by_uuid.setter
    def search_by_uuid(self, func):
        self._search_by_uuid = func

    @property
    def graph(self):
        return self._graph

    @graph.setter
    def graph(self, graph):
        self._graph = graph
        self._graph.set_vertex_properties('state', np.zeros(self._graph.nb_vertices, dtype=bool))
        self._layout = GraphLayout(self._graph)
        if self._timer is None:
            self._timer = GObject.timeout_add(TIMEOUT, self.on_timeout, self)

    def __del__(self):
        if self._timer is not None:
            GObject.source_remove(self._timer)

    def _cairo_scale(self, area, context):
        w, h = area.get_allocated_width(), area.get_allocated_height()
        positions = self._layout.positions
        min_x = np.min(positions[:, 0]) - 1
        max_x = np.max(positions[:, 0]) + 1
        min_y = np.min(positions[:, 1]) - 1
        max_y = np.max(positions[:, 1]) + 1
        s = min(w / (max_x - min_x), h / (max_y - min_y))
        context.scale(s, s)
        context.translate((w / s - min_x - max_x) / 2,
                          (h / s - min_y - max_y) / 2)

    def on_realize(self, area):
        pass

    def on_draw(self, area, context):
        if self._graph is None or self._layout is None:
            return

        context.set_source_rgb(1, 1, 1)
        context.paint()

        # Set scale transformation
        self._cairo_scale(area, context)

        # Get positions from layouter
        positions = self._layout.positions
        kind = self._graph.get_vertex_properties('kind')
        state = self._graph.get_vertex_properties('state')

        # Draw vertices
        root_color = Gdk.color_parse('lightgreen')
        does_not_exist_color = Gdk.color_parse('red')
        dependency_color = Gdk.color_parse('lightblue')
        for i, ((x, y), k, s) in enumerate(zip(positions, kind, state)):
            if k == 'root':
                context.set_source_rgb(*root_color.to_floats())
                square(context, x, y)
            elif k == 'does-not-exist':
                context.set_source_rgb(*does_not_exist_color.to_floats())
                triangle(context, x, y)
            else:
                context.set_source_rgb(*dependency_color.to_floats())
                circle(context, x, y)
            if s:
                context.fill_preserve()
                context.set_source_rgb(0, 0, 0)
                context.set_line_width(0.1)
                context.stroke()
            else:
                context.fill()

        # Draw edges
        context.set_source_rgb(0, 0, 0)
        context.set_line_width(0.1)
        for i, j in self._graph.edges:
            # Start and end position of arrow
            i_pos = positions[i].copy()
            j_pos = positions[j].copy()
            # Adjust to radius of circle
            ij = i_pos - j_pos
            normal = ij / np.linalg.norm(ij)
            perpendicular = np.array([normal[1], -normal[0]])
            i_pos -= 0.5 * normal
            j_pos += 0.5 * normal
            # Draw line
            context.move_to(*(i_pos - 0.05 * normal))
            context.line_to(*(j_pos + 0.1 * normal))
            context.stroke()
            # Draw arrow head
            context.move_to(*i_pos)
            context.line_to(*(i_pos - 0.2 * normal - 0.2 * perpendicular))
            context.line_to(*(i_pos - 0.2 * normal + 0.2 * perpendicular))
            context.fill()
            context.close_path()

    def on_motion_notify(self, area, event):
        context = area.get_window().cairo_create()
        self._cairo_scale(area, context)

        positions = self._layout.positions
        state = np.array(self._graph.get_vertex_properties('state'))
        uuids = np.array(self._graph.get_vertex_properties('uuid'))
        names = np.array(self._graph.get_vertex_properties('name'))

        cursor_pos = np.array(context.device_to_user(event.x, event.y))
        dist_sq = np.sum((positions - cursor_pos) ** 2, axis=1)

        new_state = dist_sq < 0.25
        if np.any(new_state != state):
            state = new_state
            self._graph.set_vertex_properties('state', state)

            self.queue_draw()

            if np.any(state):
                # Show popover
                positions = self._layout.positions
                x, y = positions[state][0]
                rect = Gdk.Rectangle()
                rect.x, rect.y = context.user_to_device(x, y + 0.5)
                self._popover.set_pointing_to(rect)
                self._current_uuid = uuids[state][0]
                self._popover.uuid = self._current_uuid
                self._popover.name = names[state][0]
                self._popover.show()

        if not np.any(state):
            # Hide popover if no node is active
            self._popover.hide()

    def on_show_clicked(self, user_data):
        self._popover.hide()
        # use an action to evoke search. If an according action does not exist, then nothing happens.
        search_text = dump_single_line_query_text({"uuid": self._current_uuid})
        self.get_action_group("win").activate_action('search-select-show', GLib.Variant.new_string(search_text))

    def on_timeout(self, user_data):
        try:
            self._layout.iterate()
        except Exception as e:
            logger.error(str(e))
        self.queue_draw()
        return True


GObject.type_register(DtoolGraphWidget)



#
# Copyright 2021 Johannes Hörmann
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import json


def load_query_text(text):
    return json.loads(text)


def dump_single_line_query_text(query):
    return json.dumps(query)


def dump_multi_line_query_text(query):
    return json.dumps(query, indent=4)


def single_line_sanitize_query_text(text):
    """Formats a query text coherently."""
    return dump_single_line_query_text(load_query_text(text))


def multi_line_sanitize_query_text(text):
    """Formats a query text coherently."""
    return dump_multi_line_query_text(load_query_text(text))


def is_valid_query(text):
    try:
        load_query_text(text)
    except:
        return False
    return True



#
# Copyright 2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import os

from gi.repository import Gtk


@Gtk.Template(filename=f'{os.path.dirname(__file__)}/graph_popover.ui')
class DtoolGraphPopover(Gtk.Popover):
    __gtype_name__ = 'DtoolGraphPopover'

    uuid_label = Gtk.Template.Child()
    name_label = Gtk.Template.Child()
    show_dataset_button = Gtk.Template.Child()

    def __init__(self, *args, **kwargs):
        on_show_clicked = kwargs.pop('on_show_clicked', None)
        super().__init__(*args, **kwargs)
        if on_show_clicked is not None:
            self.show_dataset_button.connect('clicked', on_show_clicked)

    @property
    def uuid(self):
        self.uuid_label.get_text()

    @uuid.setter
    def uuid(self, uuid):
        self.uuid_label.set_text(uuid)

    @property
    def name(self):
        self.name_label.get_text()

    @uuid.setter
    def name(self, name):
        self.name_label.set_text(name)



#
# Copyright 2020-2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import logging

import numpy as np
from scipy.special import erf

logger = logging.getLogger(__name__)


class SimpleGraph:
    def __init__(self):
        self._vertex_properties = []
        self._vertex1 = []
        self._vertex2 = []

    @property
    def nb_vertices(self):
        return len(self._vertex_properties)

    @property
    def vertex_properties(self):
        return self._vertex_properties

    @property
    def vertex1(self):
        return self._vertex1

    @property
    def vertex2(self):
        return self._vertex2

    @property
    def edges(self):
        return zip(self._vertex1, self._vertex2)

    @property
    def nb_edges(self):
        return len(self.vertex1)

    def add_vertex(self, **kwargs):
        self._vertex_properties += [kwargs]
        return len(self._vertex_properties) - 1

    def add_edge(self, i, j):
        nb_vertices = self.nb_vertices
        if i is None or i < 0 or i >= nb_vertices:
            raise ValueError(f'Vertex index {i} out of bounds 0 to '
                             f'{nb_vertices}.')
        if j is None or j < 0 or j >= nb_vertices:
            raise ValueError(f'Vertex index {j} out of bounds 0 to '
                             f'{nb_vertices}.')
        self._vertex1 += [i]
        self._vertex2 += [j]

    def set_vertex_properties(self, name, properties):
        for vertex, property in zip(self._vertex_properties, properties):
            vertex[name] = property

    def get_vertex_properties(self, name):
        properties = []
        for vertex in self._vertex_properties:
            properties += [vertex[name]]
        return properties


class GraphLayout:
    """
    Simple graph layouter that uses spring and electrostatic forces between
    vertices and the Fast Intertial Relaxation Engine (FIRE) for optimization.
    """

    def __init__(self, graph, spring_constant=10, equilibrium_distance=2,
                 coulomb=1, core_length=2, coulomb_exponent=1, mass=1,
                 max_timestep=1, minsteps=10, inc_timestep=1.2,
                 dec_timestep=0.5, mix=0.1, dec_mix=0.99,
                 init_iter=100):
        self.graph = graph
        self.spring_constant = spring_constant
        self.equilibrium_distance = equilibrium_distance
        self.coulomb = coulomb
        self.core_length = core_length
        self.coulomb_exponent = coulomb_exponent
        self.mass = mass
        self.max_timestep = max_timestep
        self.minsteps = minsteps
        self.inc_timestep = inc_timestep
        self.dec_timestep = dec_timestep
        self.initial_mix = mix
        self.dec_mix = dec_mix
        self.init_iter = init_iter

        self.timestep = max_timestep
        self.mix = mix
        self.cut = minsteps

        self._energy = None

        self._initialize_positions()

    @property
    def positions(self):
        return self._positions

    def _initialize_positions(self):
        nb_vertices = self.graph.nb_vertices
        n = int(np.sqrt(nb_vertices)) + 1
        grid = (np.mgrid[:n, :n].T).reshape(-1, 2)[:nb_vertices]
        self._positions = grid.astype(float) * self.equilibrium_distance
        self._velocities = np.zeros_like(self._positions)
        self._forces = np.zeros_like(self._positions)
        for i in range(self.init_iter):
            self.iterate()

    def _compute_spring_energy_and_forces(self, pos):
        nb_vertices = self.graph.nb_vertices
        if nb_vertices <= 1:
            return 0, np.zeros_like(pos)

        # Neighbor list (edge list)
        i_n = np.array(self.graph.vertex1)
        j_n = np.array(self.graph.vertex2)

        # Vertex distances
        dr_nc = pos[i_n] - pos[j_n]
        abs_dr_n = np.sqrt(np.sum(dr_nc ** 2, axis=1))

        # Energies (per pair)
        e_n = 0.5 * self.spring_constant * (
                abs_dr_n - self.equilibrium_distance) ** 2

        # Forces (per pair)
        de_n = self.spring_constant * (abs_dr_n - self.equilibrium_distance)
        df_nc = 0.5 * de_n.reshape(-1, 1) * dr_nc / abs_dr_n.reshape(-1, 1)

        # Sum for each vertex
        fx_i = np.bincount(j_n, weights=df_nc[:, 0], minlength=nb_vertices) - \
               np.bincount(i_n, weights=df_nc[:, 0], minlength=nb_vertices)
        fy_i = np.bincount(j_n, weights=df_nc[:, 1], minlength=nb_vertices) - \
               np.bincount(i_n, weights=df_nc[:, 1], minlength=nb_vertices)

        # Return energy and forces
        return np.sum(e_n), np.transpose([fx_i, fy_i])

    def _compute_coulomb_energy_and_forces(self, pos):
        nb_vertices = self.graph.nb_vertices
        if nb_vertices <= 1:
            return 0, np.zeros_like(pos)

        # Neighbor list (between all atoms)
        i_n, j_n = np.mgrid[:nb_vertices, :nb_vertices]
        i_n.shape = (-1,)
        j_n.shape = (-1,)
        m = i_n != j_n
        i_n = i_n[m]
        j_n = j_n[m]

        # Vertex distances
        dr_nc = pos[i_n] - pos[j_n]
        abs_dr_n = np.sqrt(np.sum(dr_nc ** 2, axis=1))

        # Energies (per pair)
        drnorm_n = abs_dr_n / self.core_length
        e_n = self.coulomb * erf(drnorm_n ** self.coulomb_exponent) / \
              (abs_dr_n ** self.coulomb_exponent)

        # Forces (per pair)
        de_n = self.coulomb * self.coulomb_exponent * (
                -erf(drnorm_n ** self.coulomb_exponent)
                + 2 * np.exp(-drnorm_n ** (2 * self.coulomb_exponent))
                * drnorm_n ** self.coulomb_exponent / np.sqrt(np.pi)) / \
               abs_dr_n ** (self.coulomb_exponent + 1)

        df_nc = 0.5 * de_n.reshape(-1, 1) * dr_nc / abs_dr_n.reshape(-1, 1)

        # Sum for each vertex
        fx_i = np.bincount(j_n, weights=df_nc[:, 0], minlength=nb_vertices) - \
               np.bincount(i_n, weights=df_nc[:, 0], minlength=nb_vertices)
        fy_i = np.bincount(j_n, weights=df_nc[:, 1], minlength=nb_vertices) - \
               np.bincount(i_n, weights=df_nc[:, 1], minlength=nb_vertices)

        # Return energy and forces
        return np.sum(e_n), np.transpose([fx_i, fy_i])

    def iterate(self):
        """Carry out a single step of the graph layout optimization"""

        old_energy = self._energy
        old_positions = self._positions.copy()
        old_velocities = self._velocities.copy()

        it = 0

        action = 'uphill e'
        while action == 'uphill e' and it < 5:
            # Verlet step 1
            self._velocities += 0.5 * self._forces * self.timestep / self.mass
            self._positions += self._velocities * self.timestep

            # Recompute forces
            self._energy, self._forces = self._compute_spring_energy_and_forces(
                self._positions)
            e, f = self._compute_coulomb_energy_and_forces(self._positions)
            self._energy += e
            self._forces += f

            # Verlet step 2
            self._velocities += 0.5 * self._forces * self.timestep / self.mass

            if self._energy is not None and old_energy is not None and \
                    self._energy > old_energy:
                # The energy did not decrease, decrease time step and retry!
                self._positions = old_positions.copy()
                self._velocities = old_velocities.copy()
                self.timestep *= self.dec_timestep

                action = 'uphill e'
            else:
                # Adjust velocities according to the FIRE algorithm
                v_dot_f = np.sum(self._velocities * self._forces)
                if v_dot_f < 0:
                    self._velocities = np.zeros_like(self._velocities)
                    self.cut = self.minsteps
                    self.timestep = self.timestep * self.dec_timestep
                    self.mix = self.initial_mix

                    action = 'uphill v*f'
                else:
                    v_dot_v = np.sum(self._velocities ** 2)
                    f_dot_f = np.sum(self._forces ** 2)

                    help = 0.0
                    if f_dot_f > 0:
                        help = self.mix * np.sqrt(v_dot_v / f_dot_f)

                    self._velocities = (1 - self.mix) * self._velocities + \
                                       help * self._forces

                    if self.cut < 0:
                        self.timestep = min(self.timestep * self.inc_timestep,
                                            self.max_timestep)
                        self.mix = self.mix * self.dec_mix
                    else:
                        self.cut -= 1

                    action = 'mix'

            it += 1
            # logger.debug(f'FIRE: {action}, '
            #             f'energy: {self._energy}, '
            #             f'(old energy: {old_energy}, '
            #             f'max |force|: {np.sqrt(np.max(np.sum(self._forces ** 2, axis=1)))},'
            #             f' timestep: {self.timestep}')



#
# Copyright 2021 Johannes Hörmann
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Multiprocessing utils."""

# NOTE: depending on platform, we may have to experiment with the forking methods,
# https://docs.python.org/3/library/multiprocessing.html#contexts-and-start-methods

import asyncio
import logging
import multiprocessing  # run task as child process to avoid side effects
import queue
import traceback  # forward exception from child process to parent process


logger = logging.getLogger(__name__)


def process_initializer():
    """Initialize process pool workers."""
    # Must apply patch in all worker processes.
    import dtool_lookup_gui.utils.patch

    # Avoids warning
    #   PyGIWarning: Gtk was imported without specifying a version first.
    #                Use gi.require_version('Gtk', '3.0') before import to ensure that the right version gets loaded.
    # when launching process pool. Would forking / spawning global process pool
    # before Gtk initialization be another option?
    import gi
    gi.require_version('Gtk', '3.0')


# inspired by
# https://stackoverflow.com/questions/19924104/python-multiprocessing-handling-child-errors-in-parent
class Process(multiprocessing.Process):
    """
    Class which returns child Exceptions to Parent.
    https://stackoverflow.com/a/33599967/4992248
    """

    def __init__(self, *args, **kwargs):
        multiprocessing.Process.__init__(self, *args, **kwargs)
        self._parent_conn, self._child_conn = multiprocessing.Pipe()
        self._exception = None

    def run(self):
        try:
            super().run()
            self._child_conn.send(None)
        except Exception as e:
            tb = traceback.format_exc()
            self._child_conn.send((e, tb))
            raise e  # You can still rise this exception if you need to

    @property
    def exception(self):
        if self._parent_conn.poll():
            self._exception = self._parent_conn.recv()
        return self._exception

class TargetWrapper:
    def __init__(self, target):
        self._target = target

    def __call__(self, return_value_queue, status_progress_queue, *args):
        class StatusReportClass:
            def update(status_report):
                logger.debug(f"Child process queues status report {status_report}")
                status_progress_queue.put(status_report)

        return_value_queue.put(self._target(*args, status_report_callback=StatusReportClass))

class StatusReportingChildProcessBuilder:
    """Outsource serial functions with status report handlers.

    The status report handler is expected to conform to the
    click.ProgressBar interface. In particular, it must exhibit an
    update(val) method.

    For any function that runs serial and reports status via such a callback,
    this wrapper can run them in a non-blocking forked process and forward the
    status reports via queue to the callback.

    The function must have the signature

        func(*args, status_report_callback=None)
    """
    def __init__(self, target, status_report_callback):
        self._target_wrapper = TargetWrapper(target)
        self._status_report_handler = status_report_callback

    async def __call__(self, *args):
        """Spawn child process to assure my environment stays untouched."""
        return_value_queue = multiprocessing.Queue()
        status_progress_queue = multiprocessing.Queue()
        process = Process(target=self._target_wrapper, args=[return_value_queue, status_progress_queue, *args])
        process.start()

        # wait for child to queue its return value and
        # check whether child raises exception
        while return_value_queue.empty():
            # if child raises exception, then it has terminated
            # before queueing any return value
            if process.exception:
                error, p_traceback = process.exception
                raise ChildProcessError(p_traceback)

            try:
                status_report = status_progress_queue.get_nowait()
            except queue.Empty:
                pass
            else:
                logger.debug(f"Parent process received status report {status_report}")
                self._status_report_handler.update(status_report)

            await asyncio.sleep(0.1)

        return_value = return_value_queue.get()
        # for any child that never raises an exception and does not queue
        # anything to the return_value_queue, will deadlock here
        process.join()
        return return_value


def test_function(steps, status_report_callback):
    for n in range(steps):
        print(f"Child process step {n}")
        status_report_callback.update(n)

    return True


class test_handler:
    def update(n):
        print(f"Test callback received report for step {n}")


async def test_run():
    test_process = StatusReportingChildProcessBuilder(test_function, test_handler)
    return_value = await test_process(10)
    print(f"Child process returned {return_value}.")



#
# Copyright 2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from gi.repository import GObject, Gtk


class DtoolTransferPopoverMenu(Gtk.PopoverMenu):
    __gtype_name__ = 'DtoolTransferPopoverMenu'

    _margin = 10

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, margin_top=self._margin, margin_bottom=self._margin,
                            margin_start=self._margin, margin_end=self._margin)
        self.add(self.vbox)

    def update(self, destinations, on_copy=None):
        for child in self.vbox.get_children():
            child.destroy()
        label = Gtk.Label(xalign=0, margin_top=self._margin, margin_bottom=self._margin)
        self.vbox.pack_start(label, True, False, 0)
        label.set_markup('<b>Select destination</b>')
        for destination in destinations:
            button = Gtk.ModelButton(text=destination)
            self.vbox.pack_start(button, True, False, 0)
            button.destination = destination
            if on_copy is not None:
                button.connect('clicked', on_copy)
        for child in self.get_children():
            child.show_all()


GObject.type_register(DtoolTransferPopoverMenu)



#
# Copyright 2021 Johannes Hörmann
#           2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import asyncio
import logging
import os

import jwt

from gi.repository import Gio, Gtk

from dtool_lookup_api.core.config import Config
from dtool_lookup_api.core.LookupClient import authenticate

from ..models.settings import settings
from .authentication_dialog import AuthenticationDialog
from .s3_configuration_dialog import S3ConfigurationDialog
from .smb_configuration_dialog import SMBConfigurationDialog


_DTOOL_CONFIG_PREFIXES = {
    'DTOOL_S3_ENDPOINT_': 's3',
    'DTOOL_SMB_SERVER_NAME_': 'smb',
}


logger = logging.getLogger(__name__)


def _get_base_uri(key):
    for prefix, schema in _DTOOL_CONFIG_PREFIXES.items():
        if key.startswith(prefix):
            return f'{schema}://{key[len(prefix):]}'
    return None


@Gtk.Template(filename=f'{os.path.dirname(__file__)}/settings_dialog.ui')
class SettingsDialog(Gtk.Window):
    __gtype_name__ = 'DtoolSettingsDialog'

    lookup_url_entry = Gtk.Template.Child()
    token_entry = Gtk.Template.Child()
    authenticator_url_entry = Gtk.Template.Child()
    dependency_keys_entry = Gtk.Template.Child()
    verify_ssl_certificate_switch = Gtk.Template.Child()
    base_uris_list_box = Gtk.Template.Child()

    def __init__(self, parent, **kwargs):
        super().__init__(**kwargs)

        settings.settings.bind("dependency-keys", self.dependency_keys_entry, 'text', Gio.SettingsBindFlags.DEFAULT)

        if Config.lookup_url is not None:
            self.lookup_url_entry.set_text(Config.lookup_url)
        if Config.token is not None:
            self.token_entry.set_text(Config.token)
        if Config.auth_url is not None:
            self.authenticator_url_entry.set_text(Config.auth_url)
        if Config.verify_ssl is not None:
            self.verify_ssl_certificate_switch.set_state(Config.verify_ssl)
        asyncio.create_task(self._refresh_list_of_endpoints())

    async def _refresh_list_of_endpoints(self):
        await self.base_uris_list_box.refresh(on_configure=self.on_configure_base_uri_clicked, local=False,
                                              search_results=False)

        # Plus button for adding new endpoints
        row = Gtk.ListBoxRow()
        image = Gtk.Image.new_from_icon_name('list-add-symbolic', Gtk.IconSize.BUTTON)
        image.set_margin_top(20)
        image.set_margin_bottom(20)
        image.set_margin_start(12)
        image.set_margin_end(12)
        row.connect('state-changed', self.on_base_uri_state_changed)
        row.add(image)
        self.base_uris_list_box.add(row)

        self.base_uris_list_box.show_all()

    @Gtk.Template.Callback()
    def on_delete(self, widget, event):
        # Write back configuration
        Config.lookup_url = self.lookup_url_entry.get_text()
        Config.token = self.token_entry.get_text()
        Config.auth_url = self.authenticator_url_entry.get_text()
        Config.verify_ssl = self.verify_ssl_certificate_switch.get_state()

        # Reconnect since settings may have been changed
        #asyncio.create_task(self.main_application.lookup_tab.connect())

        # Destroy window
        return False

    @Gtk.Template.Callback()
    def on_renew_token_clicked(self, widget):
        def authenticate(username, password):
            asyncio.create_task(self.retrieve_token(
                self.authenticator_url_entry.get_text(),
                username,
                password))

            # Reconnect since settings may have been changed
            #asyncio.create_task(self.main_application.lookup_tab.connect())

        AuthenticationDialog(authenticate, Config.username, Config.password).show()

    async def retrieve_token(self, auth_url, username, password):
        #self.main_application.error_bar.hide()
        try:
            token = await authenticate(auth_url, username, password)
        except Exception as e:
            logger.error(str(e))
            return
        #self.builder.get_object('token-entry').set_text(token)
        self.token_entry.set_text(token)
        await self._refresh_list_of_endpoints()

    _configuration_dialogs = {
        's3': S3ConfigurationDialog,
        'smb': SMBConfigurationDialog,
    }

    def on_configure_base_uri_clicked(self, widget):
        base_uri = widget.get_parent().get_parent().base_uri
        self._configuration_dialogs[base_uri.scheme](
            lambda: asyncio.create_task(self._refresh_list_of_endpoints()), base_uri.uri_name).show()

    def on_base_uri_state_changed(self, widget, state):
        if state == Gtk.StateType.ACTIVE:
            S3ConfigurationDialog(lambda: asyncio.create_task(self._refresh_list_of_endpoints())).show()



from PyInstaller.utils.hooks import collect_entry_point, copy_metadata

# storage brokers and their entrypoints need the following special treatment,
# as they won't be discovered by pyinstaller's default tracing mechanisms
dtool_hidden_imports = ['dtool_http', 'dtool_smb', 'dtool_s3', 'dtool_symlink']
dtool_hidden_imports_datas = []
for module in dtool_hidden_imports:
    # recursive copy_metadata needed local reinstall of
    #    $ pip install --ignore-installed six
    # in minsys2/,ingw64, would otherwise fail with
    #    RuntimeError: No metadata path found for distribution 'six'.
    dtool_hidden_imports_datas.extend(copy_metadata(module, recursive=True))

dtool_storage_brokers_datas, dtool_storage_brokers_hidden_imports = collect_entry_point("dtool.storage_brokers")

datas = [*dtool_hidden_imports_datas, *dtool_storage_brokers_datas]
hiddenimports = [*dtool_hidden_imports, *dtool_storage_brokers_hidden_imports]


#
# Copyright 2021 Lars Pastewka
#           2021 Johannes Hörmann
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import logging

from io import StringIO

from ruamel.yaml import YAML

from dtoolcore import generate_admin_metadata, generate_proto_dataset
from dtoolcore.utils import get_config_value, write_config_value_to_file, _get_config_dict_from_file, \
    generous_parse_uri, name_is_valid, NAME_VALID_CHARS_LIST
from dtool_create.dataset import _get_readme_template
from dtool_lookup_api.core.LookupClient import ConfigurationBasedLookupClient

from .datasets import DatasetModel
from .settings import settings


logger = logging.getLogger(__name__)


NON_EDITABLE_SCHEMES = ['s3']


class BaseURI:
    """Model for all base URIs"""

    def __init__(self, uri_name):
        self._uri_name = uri_name
        self._cache = None

    def __str__(self):
        return f'{self.scheme}://{self.uri_name}'

    def __eq__(self, other):
        return self.scheme == other.scheme and self.uri_name == other.uri_name

    async def all_datasets(self):
        if self._cache is None or not self._use_cache:
            self._cache = await DatasetModel.all(str(self))
        return self._cache

    # causes trouble for python < 3.9,
    # see https://docs.python.org/3.9/library/functions.html#classmethod or
    #     https://stackoverflow.com/questions/128573/using-property-on-classmethods/64738850#64738850
    # just use as per-instance property for now
    @property
    def scheme(self):
        return self._scheme

    @property
    def uri_name(self):
        return self._uri_name

    @property
    def editable(self):
        return self.scheme not in NON_EDITABLE_SCHEMES


class LocalBaseURIModel(BaseURI):
    """Model for directory on local system"""

    _scheme = 'file'
    _use_cache = False

    _local_base_uris = settings.local_base_uris

    @classmethod
    def add_directory(cls, path):
        logger.debug(f"Add local path {path} as base URI.")
        base_uri = generous_parse_uri(path)
        logger.debug(f"Parsed to qualified base URI {base_uri}.")
        if base_uri.scheme != cls._scheme:
            raise ValueError(f"The URI provided specified schema '{base_uri.scheme}', but this base URI model "
                             f"supports '{cls._scheme}'.")
        cls._local_base_uris += [base_uri.path]
        # Store such that they will reappear when restarting the program
        settings.local_base_uris = cls._local_base_uris

    @classmethod
    def remove(cls, base_uri):
        i = cls._local_base_uris.index(base_uri.uri_name)
        if i >= 0:
            del cls._local_base_uris[i]
            # Store such that they will be removed when restarting the program
            settings.local_base_uris = cls._local_base_uris

    @classmethod
    def all(cls):
        return [cls(base_uri) for base_uri in cls._local_base_uris]

    def create_dataset(self, name, readme_template_path=None):
        """Create a proto dataset."""
        # Adopted from https://github.com/jic-dtool/dtool-create/blob/master/dtool_create/dataset.py#L133

        # Check that the name is valid
        if not name_is_valid(name):
            valid_chars = " ".join(NAME_VALID_CHARS_LIST)
            raise ValueError(f"Invalid dataset name '{name}'. "
                             f"Name must be 80 characters or less. "
                             f"Dataset names may only contain the characters: {valid_chars}")

        # Administrative metadata (dtool version, etc.)
        admin_metadata = generate_admin_metadata(name)

        # Create the dataset
        base_uri = f'{self.scheme}:{self.uri_name}'
        logger.debug(f"Create dataset {name} at base URI {base_uri}.")
        proto_dataset = generate_proto_dataset(
            admin_metadata=admin_metadata,
            base_uri=base_uri)
        proto_dataset.create()  # May raise StorageBrokerOSError

        # Initialize with template README.yml
        try:
            readme_template = _get_readme_template(readme_template_path)
        except FileNotFoundError:
            # FIXME: Show an error that template could not be found
            readme_template = ''
        yaml = YAML()
        yaml.explicit_start = True
        yaml.indent(mapping=2, sequence=4, offset=2)
        descriptive_metadata = yaml.load(readme_template)
        stream = StringIO()
        yaml.dump(descriptive_metadata, stream)
        proto_dataset.put_readme(stream.getvalue())

        # Wrap in dataset model
        return DatasetModel.from_dataset(proto_dataset)


class ConfigBaseURIModel(BaseURI):
    """Model for dtool config-file based base URIs"""

    _use_cache = True

    @classmethod
    def all(cls):
        base_uris = []
        for key, value in _get_config_dict_from_file().items():
            if key.startswith(cls._prefix):
                base_uris += [cls(key[len(cls._prefix):])]
        return base_uris

    def __getattr__(self, name):
        if name.startswith('_'):
            return super().__getattr__(name)
        try:
            return get_config_value(self._properties[name].format(self._name), default='')
        except KeyError:
            raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'" )

    def __setattr__(self, name, value):
        if name.startswith('_'):
            return super().__setattr__(name, value)
        try:
            return write_config_value_to_file(self._properties[name].format(self._name), value)
        except KeyError:
            raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'" )


class S3BaseURIModel(ConfigBaseURIModel):
    """Model for managing S3 base URIs."""

    _scheme = 's3'
    _prefix = 'DTOOL_S3_ENDPOINT_'
    _properties = {
        'endpoint': 'DTOOL_S3_ENDPOINT_{}',
        'access_key_id': 'DTOOL_S3_ACCESS_KEY_ID_{}',
        'secret_access_key': 'DTOOL_S3_SECRET_ACCESS_KEY_{}',
        'dataset_prefix': 'DTOOL_S3_DATASET_PREFIX',
    }


class SMBBaseURIModel(ConfigBaseURIModel):
    """Model for managing SMB base URIs."""

    _scheme = 'smb'
    _prefix = 'DTOOL_SMB_SERVER_NAME_'
    _properties = {
        'server_name': 'DTOOL_SMB_SERVER_NAME_{}',
        'server_port': 'DTOOL_SMB_SERVER_PORT_{}',
        'domain': 'DTOOL_SMB_DOMAIN_{}',
        'service_name': 'DTOOL_SMB_SERVICE_NAME_{}',
        'path': 'DTOOL_SMB_PATH_{}',
        'username': 'DTOOL_SMB_USERNAME_{}',
        'password': 'DTOOL_SMB_PASSWORD_{}',
    }


class LookupBaseURIModel(BaseURI):
    """Model for base URIs obtained from lookup server"""

    @classmethod
    async def all(cls, username):
        async with ConfigurationBasedLookupClient() as lookup_client:
            user_info = await lookup_client.user_info(username)
        if 'search_permissions_on_base_uris' not in user_info:
            raise RuntimeError(f"Request for user '{username}' info failed, possibly not authenticated.")

        base_uris = user_info['search_permissions_on_base_uris']

        return [cls(base_uri) for base_uri in base_uris]

    def __init__(self, base_uri):
        p = generous_parse_uri(base_uri)
        self._scheme = p.scheme
        self._uri_name = p.netloc


_base_uri_models = [S3BaseURIModel, SMBBaseURIModel]


async def all(local=True, lookup=None):
    """
    Discover base URIs, either local or remote.

    Parameters
    ----------
    local : bool, optional
        List local base URIs. (Default: true)
    lookup : str, optional
        Query lookup server for additional base URIs registered with the
        lookup server but not registered locally. The `lookup` parameter
        receives the username used for authentication. (Default: None)

    Returns
    -------
    base_uris : list of :obj:`BaseURI`
        List of all discovered base URIs
    """
    base_uris = []
    for model in _base_uri_models:
        base_uris += model.all()
    if local:
        base_uris += LocalBaseURIModel.all()
    if lookup is not None:
        for base_uri in await LookupBaseURIModel.all(lookup):
            if not base_uri in base_uris:
                base_uris += [base_uri]
    return base_uris



#
# Copyright 2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

class CopyManager:
    """Keep track of running copy operations"""

    _margin = 6

    def __init__(self, progress_revealer, progress_popover):
        # Note: This is not particularly abstract, as it interacts directly with the Gtk widgets
        self._progress_revealer = progress_revealer
        self._progress_chart = progress_revealer.get_child().get_child()
        self._progress_popover = progress_popover

    async def copy(self, dataset, destination):
        self._progress_revealer.set_reveal_child(True)
        tracker = self._progress_popover.add_status_box(
            self.progress_update, f'Copying dataset »{dataset}« to »{destination}«')
        await dataset.copy(destination, progressbar=tracker)
        tracker.set_done()

        # Once all copy operations are done, we hide the pie chart and clear the popover
        if all([tracker.is_done for tracker in self._progress_popover.status_boxes]):
            for tracker in self._progress_popover.status_boxes:
                tracker.destroy()
            self._progress_revealer.set_reveal_child(False)

    def progress_update(self):
        # Refresh pie chart
        total_length = sum([len(tracker) for tracker in self._progress_popover.status_boxes])
        total_step = sum([tracker.step for tracker in self._progress_popover.status_boxes])
        if total_length > 0:
            self._progress_chart.set_fraction(total_step / total_length)
        else:
            self._progress_chart.set_fraction(1.)



import logging
import os
import socket

from dtoolcore.utils import IS_WINDOWS, windows_to_unix_path

import dtoolcore.utils

try:
    from urlparse import urlparse, urlunparse
except ImportError:
    from urllib.parse import urlparse, urlunparse

logger = logging.getLogger(__name__)

def generous_parse_uri(uri):
    """Return a urlparse.ParseResult object with the results of parsing the
    given URI. This has the same properties as the result of parse_uri.
    When passed a relative path, it determines the absolute path, sets the
    scheme to file, the netloc to localhost and returns a parse of the result.
    """
    logger.debug("In generous_pase_uri...")
    logger.debug("generous_pase_uri.input_uri: {}".format(uri))

    parse_result = urlparse(uri)

    IS_WINDOWS_DRIVE_LETTER = len(parse_result.scheme) == 1

    if parse_result.scheme == '' or IS_WINDOWS_DRIVE_LETTER:
        # ATTENTION: abspath apparently prepends the drive letter of the current working directory on Windows
        if IS_WINDOWS_DRIVE_LETTER:
            abspath = parse_result.scheme.upper() + ':' + parse_result.path
        else:
            abspath = os.path.abspath(parse_result.path)

        fixed_uri = "file://{}{}".format(
            socket.gethostname(),
            abspath
        )
        if IS_WINDOWS:
            abspath = windows_to_unix_path(abspath)
            fixed_uri = "file:///{}".format(abspath)
        parse_result = urlparse(fixed_uri)

    logger.debug("generouse_pase_uri.return: {}".format(parse_result))
    return parse_result

# as of dtoolcore v3.18.1, we have to patch dtoolcore.utils.generous_parse_uri
# to avoid a drive letter issue on windows. Otherwise, any drive letter within
# a local windows path will be replaced by the drive letter of the current
# working directory. See for example
#   $ cygpath -wa .
#   C:\msys64\home\
#   $ python -c 'import os; from urllib.parse import urlparse; parse_result = urlparse("D:/test/path"); print(os.path.abspath(parse_result.path))'
#   C:/test/path
dtoolcore.utils.generous_parse_uri = generous_parse_uri


# from https://tipsfordev.com/build-python-script-to-single-exe-with-pyinstaller
import sys
from jinja2.loaders import FileSystemLoader


class PyInstallerPackageLoader(FileSystemLoader):
    """Load templates from packages deployed as part of a Pyinstaller build.  It is constructed with
    the name of the python package and the path to the templates in that
    package::
        loader = PackageLoader('mypackage', 'views')
    If the package path is not given, ``'templates'`` is assumed.
    Per default the template encoding is ``'utf-8'`` which can be changed
    by setting the `encoding` parameter to something else.  Due to the nature
    of eggs it's only possible to reload templates if the package was loaded
    from the file system and not a zip file.
    """

    def __init__(self, package_name, package_path="templates", encoding="utf-8"):
        # Use the standard pyinstaller convention of storing additional package files
        full_package_path = f"{sys._MEIPASS}/{package_name}/{package_path}"
        # Defer everything else to the FileSystemLoader
        super().__init__(full_package_path, encoding)


def patch_jinja2_package_loader():
    """Patching the jinja2 loader which fails to locate the template when called from a pyinstaller build."""

    if getattr(sys, "frozen", False):
        import jinja2
        jinja2.PackageLoader = PyInstallerPackageLoader

patch_jinja2_package_loader()



#
# Copyright 2021 Johannes Hörmann
#           2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import locale
import logging
import math
import os
import uuid
from contextlib import contextmanager
from datetime import date, datetime
from io import StringIO

from ruamel.yaml import YAML
from ruamel.yaml.parser import ParserError
from ruamel.yaml.constructor import DuplicateKeyError
from ruamel.yaml.scanner import ScannerError

import dtoolcore

logger = logging.getLogger(__name__)


@contextmanager
def time_locale(name):
    # This code snippet was taken from:
    # https://stackoverflow.com/questions/18593661/how-do-i-strftime-a-date-object-in-a-different-locale
    saved = locale.setlocale(locale.LC_TIME)
    try:
        yield locale.setlocale(locale.LC_TIME, name)
    finally:
        locale.setlocale(locale.LC_TIME, saved)


def is_uuid(value):
    '''Check whether the data is a UUID.'''
    value = str(value)
    try:
        uuid.UUID(value)
        return True
    except ValueError:
        return False


def to_timestamp(d):
    """
    Convert a string or a timestamp to a timestamp. This is a dirty fix necessary
    because the /dataset/list route return timestamps but /dataset/search
    returns strings in older versions of the lookup server (before 0.15.0).
    """
    if type(d) is str:
        try:
            with time_locale('C'):
                d = dtoolcore.utils.timestamp(
                    datetime.strptime(d, '%a, %d %b %Y %H:%M:%S %Z'))
        except ValueError as e:
            d = -1
    return d


def datetime_to_string(d):
    return datetime.fromtimestamp(to_timestamp(d))


def date_to_string(d):
    return date.fromtimestamp(to_timestamp(d))




def fill_readme_tree_store(store, data, parent=None):
    def append_entry(store, entry, value, parent):
        # Check whether the data is a UUID. We then enable a
        # hyperlink-like navigation between datasets
        is_u = is_uuid(value)
        if is_u:
            markup = '<span foreground="blue" underline="single">' \
                     f'{str(value)}</span>'
        else:
            markup = f'<span>{str(value)}</span>'
        store.append(parent,
                     [entry, str(value), is_u, markup])

    def fill_readme_tree_store_from_list(store, list_data, parent=None):
        for i, current_data in enumerate(list_data):
            entry = f'{i + 1}'
            if isinstance(current_data, list):
                current_parent = store.append(parent,
                                              [entry, None, False, None])
                fill_readme_tree_store_from_list(store, current_data,
                                                 parent=current_parent)
            elif isinstance(current_data, dict):
                current_parent = store.append(parent,
                                              [entry, None, False, None])
                fill_readme_tree_store(store, current_data,
                                       parent=current_parent)
            else:
                append_entry(store, entry, current_data, parent)

    if data is not None:
        for entry, value in data.items():
            if isinstance(value, list):
                current = store.append(parent,
                                       [entry, None, False, None])
                fill_readme_tree_store_from_list(store, value, parent=current)
            elif isinstance(value, dict):
                current = store.append(parent,
                                       [entry, None, False, None])
                fill_readme_tree_store(store, value, parent=current)
            else:
                append_entry(store, entry, value, parent)


def _validate_readme(readme_content):
    """Return (YAML string, error message)."""
    yaml = YAML()
    # Ensure that the content is valid YAML.
    try:
        readme_formatted = yaml.load(readme_content)
        return readme_formatted, None
    except (ParserError, DuplicateKeyError, ScannerError) as message:
        return None, str(message)


def _standardize_readme(readme_content):
    # Create YAML object to standardise the output formatting.
    yaml = YAML()
    yaml.explicit_start = True
    yaml.indent(mapping=2, sequence=4, offset=2)

    # Validate the YAML.
    readme_formatted, error = _validate_readme(readme_content)
    if error is not None:
        raise ValueError(error)

    stream = StringIO()
    yaml.dump(readme_formatted, stream)
    return stream.getvalue()



#
# Copyright 2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import os

from gi.repository import Gio


class Settings:
    def __init__(self):
        schema_source = Gio.SettingsSchemaSource.new_from_directory(
            os.path.abspath(f'{os.path.dirname(__file__)}/..'), Gio.SettingsSchemaSource.get_default(), False)
        schema = Gio.SettingsSchemaSource.lookup(
            schema_source, "de.uni-freiburg.dtool-lookup-gui", False)
        self.settings = Gio.Settings.new_full(schema, None, None)

    @property
    def dependency_keys(self):
        return self.settings.get_string('dependency-keys')

    @property
    def local_base_uris(self):
        return self.settings.get_strv('local-base-uris')

    @local_base_uris.setter
    def local_base_uris(self, value):
        self.settings.set_strv('local-base-uris', value)

settings = Settings()



#
# Copyright 2020-2021 Lars Pastewka
#           2020-2021 Johannes Hörmann
#           2020 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from subprocess import Popen

from setuptools import setup
from setuptools.command.install import install


class CustomInstallCommand(install):
    def run(self):
        install.run(self)
        Popen("glib-compile-schemas " + self.install_lib + "/dtool_lookup_gui/",
              shell=True).wait()


url = 'https://github.com/IMTEK-Simulation/dtool-lookup-gui'
readme = open('README.rst', encoding="utf8").read()

setup(
    cmdclass={
        'install': CustomInstallCommand,
    },
    name='dtool-lookup-gui',
    packages=['dtool_lookup_gui'],
    description='Graphical user interface for dtool',
    long_description=readme,
    include_package_data=True,
    author='Lars Pastewka',
    author_email='lars.pastewka@imtek.uni-freiburg.de',
    url=url,
    use_scm_version=True,
    install_requires=[
        'dtoolcore>=3.17',
        'dtool-create>=0.23.4',
        'dtool-info>=0.16.2',
        'dtool-lookup-api>=0.5',
        'aiohttp>=3.6',
        'gbulb>=0.6',
        'pyyaml>=5.3',
        'ruamel.yaml',
        'PyGObject>=3.36',
        'scipy>=1.5',
        'numpy',
        'jwt'
    ],
    setup_requires=[
        'setuptools_scm>=3.5.0'
    ],
    license='MIT'
)



#
# Copyright 2021 Lars Pastewka
#           2021 Johannes Hörmann
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import asyncio
import logging
import os
import shutil
import traceback
import urllib.parse
from functools import reduce

from gi.repository import Gio, GLib, Gtk, GtkSource

import dtoolcore.utils
from dtool_info.utils import sizeof_fmt

import dtool_lookup_api.core.config
from dtool_lookup_api.core.LookupClient import ConfigurationBasedLookupClient
# As of dtool-lookup-api 0.5.0, the following line still is a necessity to
# disable prompting for credentials on the command line. This behavior
# will change in future versions.
dtool_lookup_api.core.config.Config.interactive = False

from ..models.base_uris import all, LocalBaseURIModel
from ..models.datasets import DatasetModel
from ..models.settings import settings
from ..utils.copy_manager import CopyManager
from ..utils.date import date_to_string
from ..utils.dependency_graph import DependencyGraph
from ..utils.logging import FormattedSingleMessageGtkInfoBarHandler, DefaultFilter
from ..utils.query import (is_valid_query, dump_single_line_query_text)
from ..widgets.base_uri_row import DtoolBaseURIRow
from ..widgets.search_popover import DtoolSearchPopover
from ..widgets.search_results_row import DtoolSearchResultsRow
from .dataset_name_dialog import DatasetNameDialog
from .settings_dialog import SettingsDialog
from .log_window import LogWindow

_logger = logging.getLogger(__name__)


def _fill_manifest_tree_store(store, manifest, parent=None):
    nodes = {}

    store.clear()

    def find_or_create_parent_node(path, top_parent):
        if not path:
            return top_parent
        try:
            return nodes[path]
        except KeyError:
            head, tail = os.path.split(path)
            parent = find_or_create_parent_node(head, top_parent)
            new_node = store.append(parent, [tail, '', '', ''])
            nodes[path] = new_node
            return new_node

    for uuid, values in sorted(manifest, key=lambda kv: kv[1]['relpath']):
        head, tail = os.path.split(values['relpath'])
        store.append(find_or_create_parent_node(head, parent),
                     [tail,
                      sizeof_fmt(values['size_in_bytes']).strip(),
                      f'{date_to_string(values["utc_timestamp"])}',
                      uuid])


@Gtk.Template(filename=f'{os.path.dirname(__file__)}/main_window.ui')
class MainWindow(Gtk.ApplicationWindow):
    __gtype_name__ = 'DtoolMainWindow'

    _max_nb_datasets = 100

    create_dataset_button = Gtk.Template.Child()
    menu_button = Gtk.Template.Child()

    search_entry = Gtk.Template.Child()

    #copy_dataset_spinner = Gtk.Template.Child()

    base_uri_list_box = Gtk.Template.Child()
    dataset_list_box = Gtk.Template.Child()

    main_stack = Gtk.Template.Child()
    main_paned = Gtk.Template.Child()
    main_label = Gtk.Template.Child()
    main_spinner = Gtk.Template.Child()

    dataset_stack = Gtk.Template.Child()
    dataset_box = Gtk.Template.Child()
    dataset_label = Gtk.Template.Child()

    uuid_label = Gtk.Template.Child()
    uri_label = Gtk.Template.Child()
    name_label = Gtk.Template.Child()
    created_by_label = Gtk.Template.Child()
    frozen_at_label = Gtk.Template.Child()
    size_label = Gtk.Template.Child()

    show_button = Gtk.Template.Child()
    add_items_button = Gtk.Template.Child()
    freeze_button = Gtk.Template.Child()
    copy_button = Gtk.Template.Child()

    progress_revealer = Gtk.Template.Child()
    progress_button = Gtk.Template.Child()
    progress_popover = Gtk.Template.Child()

    edit_readme_switch = Gtk.Template.Child()
    save_metadata_button = Gtk.Template.Child()

    dependency_stack = Gtk.Template.Child()
    dependency_view = Gtk.Template.Child()
    dependency_spinner = Gtk.Template.Child()
    dependency_graph_widget = Gtk.Template.Child()

    readme_source_view = Gtk.Template.Child()
    readme_spinner = Gtk.Template.Child()
    readme_stack = Gtk.Template.Child()
    readme_view = Gtk.Template.Child()

    manifest_spinner = Gtk.Template.Child()
    manifest_stack = Gtk.Template.Child()
    manifest_tree_view = Gtk.Template.Child()
    manifest_tree_store = Gtk.Template.Child()
    manifest_view = Gtk.Template.Child()

    settings_button = Gtk.Template.Child()

    error_bar = Gtk.Template.Child()
    error_label = Gtk.Template.Child()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.application = self.get_application()

        self.main_stack.set_visible_child(self.main_label)
        self.dataset_stack.set_visible_child(self.dataset_label)

        self.readme_buffer = self.readme_source_view.get_buffer()
        lang_manager = GtkSource.LanguageManager()
        self.readme_buffer.set_language(lang_manager.get_language("yaml"))
        self.readme_buffer.set_highlight_syntax(True)
        self.readme_buffer.set_highlight_matching_brackets(True)

        self.error_bar.set_revealed(False)
        self.progress_revealer.set_reveal_child(False)

        # connect log handler to error bar
        root_logger = logging.getLogger()
        self.log_handler = FormattedSingleMessageGtkInfoBarHandler(info_bar=self.error_bar, label=self.error_label)
        # exclude unwanted log messages from being displayed in error bar
        self.log_handler.addFilter(DefaultFilter())
        root_logger.addHandler(self.log_handler)

        # connect a search popover with search entry
        self.search_popover = DtoolSearchPopover(search_entry=self.search_entry)
        self.log_window = LogWindow(application=self.application)

        # window-scoped actions

        # search action
        search_text_variant = GLib.Variant.new_string("dummy")
        search_action = Gio.SimpleAction.new("search", search_text_variant.get_type())
        search_action.connect("activate", self.do_search)
        self.add_action(search_action)

        # select row by row index in dataset list box action
        row_index_variant = GLib.Variant.new_uint32(0)
        select_dataset_action = Gio.SimpleAction.new("select-dataset", row_index_variant.get_type())
        select_dataset_action.connect("activate", self.do_select_dataset_row_by_row_index)
        self.add_action(select_dataset_action)

        # select row by uri in dataset list box action
        uri_variant = GLib.Variant.new_string('dummy')
        select_dataset_by_uri_action = Gio.SimpleAction.new("select-dataset-by-uri", uri_variant.get_type())
        select_dataset_by_uri_action.connect("activate", self.do_select_dataset_row_by_uri)
        self.add_action(select_dataset_by_uri_action)

        # show details of dataset by row index in dataset list box action
        row_index_variant = GLib.Variant.new_uint32(0)
        show_dataset_action = Gio.SimpleAction.new("show-dataset", row_index_variant.get_type())
        show_dataset_action.connect("activate", self.do_show_dataset_details_by_row_index)
        self.add_action(show_dataset_action)

        # show details of dataset by uri in dataset list box action
        uri_variant = GLib.Variant.new_string("dummy")
        show_dataset_by_uri_action = Gio.SimpleAction.new("show-dataset-by-uri", uri_variant.get_type())
        show_dataset_by_uri_action.connect("activate", self.do_show_dataset_details_by_uri)
        self.add_action(show_dataset_by_uri_action)

        # search, select and show first search result subsequently
        row_index_variant = GLib.Variant.new_string("dummy")
        search_select_show_action = Gio.SimpleAction.new("search-select-show", row_index_variant.get_type())
        search_select_show_action.connect("activate", self.do_search_select_and_show)
        self.add_action(search_select_show_action)

        # get item
        dest_file_variant = GLib.Variant.new_string("dummy")
        get_item_action = Gio.SimpleAction.new("get-item", dest_file_variant.get_type())
        get_item_action.connect("activate", self.do_get_item)
        self.add_action(get_item_action)

        self.dependency_graph_widget.search_by_uuid = self._search_by_uuid

        self._copy_manager = CopyManager(self.progress_revealer, self.progress_popover)

        _logger.debug(f"Constructed main window for app '{self.application.get_application_id()}'")

    # utility methods
    def refresh(self):
        asyncio.create_task(self.base_uri_list_box.refresh())

    # removed these utility functions from inner scope of on_search_activate
    # in order to decouple actual signal handler and functionality
    def _update_search_summary(self, datasets):
        row = self.base_uri_list_box.search_results_row
        total_size = sum([0 if dataset.size_int is None else dataset.size_int for dataset in datasets])
        row.info_label.set_text(f'{len(datasets)} datasets, {sizeof_fmt(total_size).strip()}')

    async def _fetch_search_results(self, keyword, on_show=None):
        row = self.base_uri_list_box.search_results_row
        row.start_spinner()

        try:
            # datasets = await DatasetModel.search(keyword)
            if keyword:
                if is_valid_query(keyword):
                    _logger.debug("Valid query specified.")
                    datasets = await DatasetModel.query(keyword)
                else:
                    _logger.debug("Specified search text is not a valid query, just perform free text search.")
                    # NOTE: server side allows a dict with the key-value pairs
                    # "free_text", "creator_usernames", "base_uris", "uuids", "tags",
                    # via route '/dataset/search', where all except "free_text"
                    # can be lists and are translated to logical "and" or "or"
                    # constructs on the server side. With the special treatment
                    # of the 'uuid' keyword above, should we introduce similar
                    # options for the other available keywords?
                    datasets = await DatasetModel.search(keyword)
            else:
                _logger.debug("No keyword specified, list all datasets.")
                datasets = await DatasetModel.query_all()

            if len(datasets) > self._max_nb_datasets:
                _logger.warning(
                    f"{len(datasets)} search results exceed allowed displayed maximum of {self._max_nb_datasets}. "
                    f"Only the first {self._max_nb_datasets} results are shown. Narrow down your search.")
            datasets = datasets[:self._max_nb_datasets]  # Limit number of datasets that are shown
            row.search_results = datasets  # Cache datasets
            self._update_search_summary(datasets)
            if self.base_uri_list_box.get_selected_row() == row:
                # Only update if the row is still selected
                self.dataset_list_box.fill(datasets, on_show=on_show)

        except Exception as e:
            self.show_error(e)

        self.base_uri_list_box.select_search_results_row()
        self.main_stack.set_visible_child(self.main_paned)
        row.stop_spinner()

    def _search_by_uuid(self, uuid):
        search_text = dump_single_line_query_text({"uuid": uuid})
        self._search_by_search_text(search_text)

    def _search_by_search_text(self, search_text):
        self.activate_action('search-select-show', GLib.Variant.new_string(search_text))

    # utility methods - dataset selection
    def _select_dataset_row_by_row_index(self, index):
        """Select dataset row in dataset list box by index."""
        row = self.dataset_list_box.get_row_at_index(index)
        if row is not None:
            _logger.debug(f"{row} selected.")
            self.dataset_list_box.select_row(row)
        else:
            _logger.info(f"No row with index {index} available for selection.")

    def _select_dataset_row_by_uri(self, uri):
        """Select dataset row in dataset list box by uri."""
        index = self.dataset_list_box.get_row_index_from_uri(uri)
        self._select_dataset_row_by_row_index(index)

    def _show_dataset_details(self, dataset):
        asyncio.create_task(self._update_dataset_view(dataset))
        self.dataset_stack.set_visible_child(self.dataset_box)

    def _show_dataset_details_by_row_index(self, index):
        row = self.dataset_list_box.get_row_at_index(index)
        if row is not None:
            _logger.debug(f"{row.dataset.name} shown.")
            self._show_dataset_details(row.dataset)
        else:
            _logger.info(f"No row with index {index} available for selection.")

    def _show_dataset_details_by_uri(self, uri):
        """Select dataset row in dataset list box by uri."""
        index = self.dataset_list_box.get_row_index_from_uri(uri)
        self._show_dataset_details_by_row_index(index)

    def _select_and_show_by_row_index(self, index=0):
        self._select_dataset_row_by_row_index(index)
        self._show_dataset_details_by_row_index(index)

    def _select_and_show_by_uri(self, uri):
        self._select_dataset_row_by_uri(uri)
        self._show_dataset_details_by_uri(uri)

    def _search(self, search_text, on_show=None):
        _logger.debug(f"Evoke search with search text {search_text}.")
        self.main_stack.set_visible_child(self.main_spinner)
        row = self.base_uri_list_box.search_results_row
        row.search_results = None
        asyncio.create_task(self._fetch_search_results(search_text, on_show))

    def _search_select_and_show(self, search_text):
        _logger.debug(f"Search '{search_text}'...")
        self._search(search_text, on_show=lambda _: self._select_and_show_by_row_index())

    def _get_selected_items(self):
        """Returns (name uuid) tuples of items selected in manifest tree store."""
        selection = self.manifest_tree_view.get_selection()
        model, paths = selection.get_selected_rows()

        items = []
        for path in paths:
            column_iter = model.get_iter(path)
            item_name = model.get_value(column_iter, 0)
            item_uuid = model.get_value(column_iter, 3)
            items.append((item_name, item_uuid))

        return items

    # actions
    def do_search(self, action, value):
        """Evoke search tas for specific search text."""
        search_text = value.get_string()
        self._search(search_text)

    def do_select_dataset_row_by_row_index(self, action, value):
        """Select dataset row by index."""
        row_index = value.get_uint32()
        self._select_dataset_row_by_row_index(row_index)

    def do_select_dataset_row_by_uri(self, action, value):
        """Select dataset row by uri."""
        uri = value.get_string()
        self._select_dataset_row_by_uri(uri)

    def do_show_dataset_details_by_row_index(self, action, value):
        row_index = value.get_uint32()
        self._show_dataset_details_by_row_index(row_index)

    def do_show_dataset_details_by_uri(self, action, value):
        uri = value.get_string()
        self._show_dataset_details_by_uri(uri)

    def do_search_select_and_show(self, action, value):
        """Evoke search task for specific search text, select and show 1st row of resuls subsequntly."""
        search_text = value.get_string()
        self._search_select_and_show(search_text)

    def do_get_item(self, action, value):
        """Copy currently selected manifest item in currently selected dataset to specified destination."""
        dest_file = value.get_string()

        dataset = self.dataset_list_box.get_selected_row().dataset

        items = self._get_selected_items()
        if len(items) !=1:
            raise ValueError("Can only get one item at a time.")
        item_name, item_uuid = items[0]

        async def _get_item(dataset, item_uuid):
            cached_file = await dataset.get_item(item_uuid)
            shutil.copyfile(cached_file, dest_file)

        asyncio.create_task(_get_item(dataset, item_uuid))

    # signal handlers
    @Gtk.Template.Callback()
    def on_settings_clicked(self, widget):
        SettingsDialog(self).show()

    @Gtk.Template.Callback()
    def on_logging_clicked(self, widget):
        self.log_window.show()

    @Gtk.Template.Callback()
    def on_base_uri_selected(self, list_box, row):
        if row is None:
            # this callback apparently gets evoked with row=None when an entry is deleted / unselected (?) from the base URI list
            return

        def update_base_uri_summary(datasets):
            total_size = sum([0 if dataset.size_int is None else dataset.size_int for dataset in datasets])
            row.info_label.set_text(f'{len(datasets)} datasets, {sizeof_fmt(total_size).strip()}')

        async def _select_base_uri():
            row.start_spinner()

            if isinstance(row, DtoolBaseURIRow):
                try:
                    _logger.debug(f"Selected base URI {row.base_uri}.")
                    datasets = await row.base_uri.all_datasets()
                    _logger.debug(f"Found {len(datasets)} datasets.")
                    update_base_uri_summary(datasets)
                    if self.base_uri_list_box.get_selected_row() == row:
                        # Only update if the row is still selected
                        self.dataset_list_box.fill(datasets)
                except Exception as e:
                    self.show_error(e)
                self.main_stack.set_visible_child(self.main_paned)
            elif isinstance(row, DtoolSearchResultsRow):
                _logger.debug("Selected search results.")
                # This is the search result
                if row.search_results is not None:
                    _logger.debug(f"Fill dataset list with {len(row.search_results)} search results.")
                    self.dataset_list_box.fill(row.search_results)
                    self.main_stack.set_visible_child(self.main_paned)
                else:
                    _logger.debug("No search results cached (likely first activation after app startup).")
                    _logger.debug("Mock emit search_entry activate signal once.")
                    self.main_stack.set_visible_child(self.main_label)
                    self.search_entry.emit("activate")
            else:
                raise TypeError(f"Handling of {type(row)} not implemented.")

            row.stop_spinner()
            row.task = None

        self.main_stack.set_visible_child(self.main_spinner)
        self.create_dataset_button.set_sensitive(not isinstance(row, DtoolSearchResultsRow) and
                                                 row.base_uri.editable)
        if row.task is None:
            _logger.debug("Spawn select_base_uri task.")
            row.task = asyncio.create_task(_select_base_uri())

    @Gtk.Template.Callback()
    def on_search_activate(self, widget):
        """Search activated (usually by hitting Enter after typing in the search entry)."""
        search_text = self.search_entry.get_text()
        self._search_by_search_text(search_text)

    @Gtk.Template.Callback()
    def on_dataset_selected(self, list_box, row):
        if row is not None:
            row_index = row.get_index()
            _logger.debug(f"Selected row {row_index}.")
            self.activate_action('show-dataset', GLib.Variant.new_uint32(row_index))

    @Gtk.Template.Callback()
    def on_open_local_directory_clicked(self, widget):
        # File chooser dialog (select directory)
        dialog = Gtk.FileChooserDialog(
            title="Open local directory",
            parent=self,
            action=Gtk.FileChooserAction.SELECT_FOLDER
        )
        dialog.add_buttons(
            Gtk.STOCK_CANCEL,
            Gtk.ResponseType.CANCEL,
            Gtk.STOCK_OPEN,
            Gtk.ResponseType.OK,
        )

        # Attention: Avoid run method!
        # Unlike GLib, Python does not support running the EventLoop recursively.
        # Gbulb uses the GLib event loop, hence this works. If we move to another
        # implementation (e.g. https://gitlab.gnome.org/GNOME/pygobject/-/merge_requests/189)
        # that uses the asyncio event loop this will break.
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            # Quote from https://athenajc.gitbooks.io/python-gtk-3-api/content/gtk-group/gtkfilechooser.html:
            #
            # When the user is finished selecting files in a Gtk.FileChooser, your program can get the selected names
            # either as filenames or as URIs. For URIs, the normal escaping rules are applied if the URI contains
            # non-ASCII characters.
            #
            # However, filenames are always returned in the character set specified by the G_FILENAME_ENCODING
            # environment variable.
            #
            # This means that while you can pass the result of Gtk.FileChooser::get_filename() to open() or fopen(),
            # you may not be able to directly set it as the text of a Gtk.Label widget unless you convert it first to
            # UTF-8, which all GTK+ widgets expect. You should use g_filename_to_utf8() to convert filenames into
            # strings that can be passed to GTK+ widgets.
            uri, = dialog.get_uris()
            # For using URI scheme on local paths, we have to unquote characters to be
            uri = urllib.parse.unquote(uri, encoding='utf-8', errors='replace')
            # Add directory to local inventory
            LocalBaseURIModel.add_directory(uri)
        elif response == Gtk.ResponseType.CANCEL:
            uri = None
        dialog.destroy()

        # Refresh view of base URIs
        asyncio.create_task(self.base_uri_list_box.refresh())

    @Gtk.Template.Callback()
    def on_create_dataset_clicked(self, widget):
        DatasetNameDialog(on_confirmation=self._create_dataset).show()

    @Gtk.Template.Callback()
    def on_show_clicked(self, widget):
        Gio.AppInfo.launch_default_for_uri(str(self.dataset_list_box.get_selected_row().dataset))

    @Gtk.Template.Callback()
    def on_add_items_clicked(self, widget):
        dialog = Gtk.FileChooserDialog(
            title="Add items", parent=self,
            action=Gtk.FileChooserAction.OPEN
        )
        dialog.add_buttons(
            Gtk.STOCK_CANCEL,
            Gtk.ResponseType.CANCEL,
            Gtk.STOCK_OPEN,
            Gtk.ResponseType.OK,
        )
        dialog.set_select_multiple(True)

        # Attention: Avoid run method!
        # Unlike GLib, Python does not support running the EventLoop recursively.
        # Gbulb uses the GLib event loop, hence this works. If we move to another
        # implementation (e.g. https://gitlab.gnome.org/GNOME/pygobject/-/merge_requests/189)
        # that uses the asyncio event loop this will break.
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            uris = dialog.get_uris()
            fpaths = dialog.get_filenames()
            for fpath in fpaths:
                # uri = urllib.parse.unquote(uri, encoding='utf-8', errors='replace')
                self._add_item(fpath)
        elif response == Gtk.ResponseType.CANCEL:
            pass
        dialog.destroy()

    @Gtk.Template.Callback()
    def on_manifest_row_activated(self, tree_view, path, column):
        """Handler for "row-activated" signal.

        Signal emitted when the method gtk-tree-view-row-activated is called or the user double clicks a treeview row.
        It is also emitted when a non-editable row is selected and one of the keys: Space, Shift+Space, Return or Enter
        is pressed. (https://www.gnu.org/software/guile-gnome/docs/gtk/html/GtkTreeView.html)"""

        items = self._get_selected_items()
        if len(items) !=1:
            raise ValueError("Can only get one item at a time.")
        item_name, item_uuid = items[0]
        self._show_get_item_dialog(item_name, item_uuid)

    @Gtk.Template.Callback()
    def on_edit_readme_state_set(self, widget, state):
        self.readme_source_view.set_editable(state)
        self.save_metadata_button.set_sensitive(state)
        if not state:
            # We need to save the metadata
            text = self.readme_buffer.get_text(self.readme_buffer.get_start_iter(),
                                               self.readme_buffer.get_end_iter(),
                                               False)
            try:
                self.dataset_list_box.get_selected_row().dataset.put_readme(text)
            except Exception as e:
                self.show_error(e)

    @Gtk.Template.Callback()
    def on_save_metadata_button_clicked(self, widget):
        self.edit_readme_switch.set_state(False)

    @Gtk.Template.Callback()
    def on_freeze_clicked(self, widget):
        row = self.dataset_list_box.get_selected_row()
        dialog = Gtk.MessageDialog(self, Gtk.DialogFlags.MODAL, Gtk.MessageType.QUESTION, Gtk.ButtonsType.OK_CANCEL,
                                   f'You are about to freeze dataset "{row.dataset.name}". Items can no longer be '
                                   'added, removed or modified after freezing the dataset. (You will still be able to '
                                   'edit the metadata README.yml.) Please confirm freezing of this dataset.')
        # Attention: Avoid run method!
        # Unlike GLib, Python does not support running the EventLoop recursively.
        # Gbulb uses the GLib event loop, hence this works. If we move to another
        # implementation (e.g. https://gitlab.gnome.org/GNOME/pygobject/-/merge_requests/189)
        # that uses the asyncio event loop this will break.
        response = dialog.run()
        dialog.destroy()
        if response == Gtk.ResponseType.OK:
            uri = row.dataset.uri  # URI won't change in freeze process
            row.freeze()
            self.dataset_list_box.show_all()
            self.get_action_group("win").activate_action('select-dataset-by-uri', GLib.Variant.new_string(uri))
            self.get_action_group("win").activate_action('show-dataset-by-uri', GLib.Variant.new_string(uri))

    @Gtk.Template.Callback()
    def on_error_bar_close(self, widget):
        _logger.debug("Hide error bar.")
        self.error_bar.set_revealed(False)

    @Gtk.Template.Callback()
    def on_error_bar_response(self, widget, response_id):
        if response_id == Gtk.ResponseType.CLOSE:
            self.error_bar.set_revealed(False)

    def on_copy_clicked(self, widget):
        async def _copy():
            try:
                await self._copy_manager.copy(self.dataset_list_box.get_selected_row().dataset, widget.destination)
            except Exception as e:
                self.show_error(e)

        asyncio.create_task(_copy())

    def _show_get_item_dialog(self, item_name, item_uuid):
        dialog = Gtk.FileChooserDialog(
            title=f"Download item {item_uuid}: {item_name}", parent=self,
            action=Gtk.FileChooserAction.SAVE
        )
        dialog.add_buttons(
            Gtk.STOCK_CANCEL,
            Gtk.ResponseType.CANCEL,
            Gtk.STOCK_OK,
            Gtk.ResponseType.OK,
        )
        dialog.set_current_name(item_name)
        dialog.set_do_overwrite_confirmation(True)

        # Attention: Avoid run method!
        # Unlike GLib, Python does not support running the EventLoop recursively.
        # Gbulb uses the GLib event loop, hence this works. If we move to another
        # implementation (e.g. https://gitlab.gnome.org/GNOME/pygobject/-/merge_requests/189)
        # that uses the asyncio event loop this will break.
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            dest_file = dialog.get_filename()
            self.activate_action('get-item', GLib.Variant.new_string(dest_file))
        elif response == Gtk.ResponseType.CANCEL:
            pass
        dialog.destroy()

    # TODO: move to the model
    def _add_item(self, fpath):
        handle = os.path.basename(fpath)
        dataset = dtoolcore.ProtoDataSet.from_uri(self.dataset_list_box.get_selected_row().dataset.uri)
        dataset.put_item(fpath, handle)

    def _create_dataset(self, name):
        base_uri = self.base_uri_list_box.get_selected_row()
        if base_uri is not None:
            self.dataset_list_box.add_dataset(base_uri.base_uri.create_dataset(name))
            self.dataset_list_box.show_all()

    async def _update_dataset_view(self, dataset):
        _logger.debug("In _update_dataset_view.")

        self.uuid_label.set_text(dataset.uuid)
        self.uri_label.set_text(dataset.uri)
        self.name_label.set_text(dataset.name)
        self.created_by_label.set_text(dataset.creator)
        self.frozen_at_label.set_text(dataset.date)
        self.size_label.set_text(dataset.size_str.strip())

        # This binary distinction will allow manipulation of all datasets via
        # the according StorageBroker, as long as latter implements the
        # desired functionality
        if dataset.type == 'lookup':
            _logger.debug("Any other dataset access")
            self.show_button.set_sensitive(False)
            self.add_items_button.set_sensitive(False)
            self.freeze_button.set_sensitive(False)
            self.copy_button.set_sensitive(True)
        # if necessary, insert scheme-specific clauses later, i.e.
        # elif dataset.scheme == 's3', ....
        else:  # per default, treat all endpoints equally here
            _logger.debug("File system dataset access")
            self.show_button.set_sensitive(True)
            self.add_items_button.set_sensitive(not dataset.is_frozen)
            self.freeze_button.set_sensitive(not dataset.is_frozen)
            self.copy_button.set_sensitive(dataset.is_frozen)

        async def _get_readme():
            self.readme_stack.set_visible_child(self.readme_spinner)
            self.readme_buffer.set_text(await dataset.get_readme())
            self.readme_stack.set_visible_child(self.readme_view)

        async def _get_manifest():
            self.manifest_stack.set_visible_child(self.manifest_spinner)
            _fill_manifest_tree_store(self.manifest_tree_store, await dataset.get_manifest())
            self.manifest_stack.set_visible_child(self.manifest_view)

        _logger.debug("Get readme.")
        asyncio.create_task(_get_readme())
        _logger.debug("Get manifest.")
        asyncio.create_task(_get_manifest())

        if dataset.type == 'lookup':
            self.dependency_stack.show()
            _logger.debug("Selected dataset is lookup result.")
            asyncio.create_task(self._compute_dependencies(dataset))
        else:
            _logger.debug("Selected dataset is accessed directly.")
            self.dependency_stack.hide()

        await self._update_copy_button(dataset)

    async def _update_copy_button(self, selected_dataset):
        destinations = []
        for base_uri in await all():
            if str(base_uri) != str(selected_dataset.base_uri):
                destinations += [str(base_uri)]
        self.copy_button.get_popover().update(destinations, self.on_copy_clicked)

    async def _compute_dependencies(self, dataset):
        _logger.debug("Compute dependencies for dataset '{dataset.uuid}'.")
        self.dependency_stack.set_visible_child(self.dependency_spinner)

        # Compute dependency graph
        dependency_graph = DependencyGraph()
        async with ConfigurationBasedLookupClient() as lookup:
            _logger.debug("Wait for depenedency graph for '{dataset.uuid}' queried from lookup server.")
            await dependency_graph.trace_dependencies(lookup, dataset.uuid, dependency_keys=settings.dependency_keys)

        # Show message if uuids are missing
        missing_uuids = dependency_graph.missing_uuids
        if missing_uuids:
            _logger.warning('The following UUIDs were found during dependency graph calculation but are not present '
                            'in the database: {}'.format(reduce(lambda a, b: a + ', ' + b, missing_uuids)))

        self.dependency_graph_widget.graph = dependency_graph.graph
        self.dependency_stack.set_visible_child(self.dependency_view)

    def show_error(self, exception):
        _logger.error(traceback.format_exc())


#
# Copyright 2021 Lars Pastewka
#           2021 Johannes Hörmann
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import logging
import re
from typing import Optional

from gi.repository import Gtk, Pango

DEFAULT_FORMATTER = logging.Formatter('%(asctime)s:%(name)s:%(levelname)s: %(message)s')
SINGLE_MESSAGE_FORMATTER = logging.Formatter('%(levelname)s: %(message)s')

DEFAULT_TEXT_BUFFER_MAX_LINES = 1000
DEFAULT_ENTRY_MAX_LINES = 5
DEFAULT_LABEL_MAX_LINES = 1

# filters

# We don't want certain exceptions to clutter the logs in production mode
DEFAULT_EXCLUDE_PATTERNS = [
    'WinError 10053',  # ConnectionAbortedError: [WinError 10053] An established connection was aborted by the software in your host machine
    'Fatal read error on pipe transport',  # apparently arises in connection with above [WinError 10053],
    'client_session: <aiohttp.client.ClientSession object at .*>',  # apparently arises in connection with 'RuntimeError: Authentication failed'
    'Unclosed client session',  # apparently arises in connection with 'RuntimeError: Authentication failed',
]

DEFAULT_EXCLUDE_REGEX = [
    re.compile(pattern) for pattern in DEFAULT_EXCLUDE_PATTERNS
]

class DefaultFilter(logging.Filter):
    def filter(self, record):
        """Filter out any log messages that match any of the defined exclusion patterns."""
        for regex in DEFAULT_EXCLUDE_REGEX:
            if regex.search(record.msg) is not None:
                return False

        return True

# formatter mixins

class FormattedHandlerMixin():
    """Mixin for logging.Handler derivatives. Assigns formatter at creation."""
    def __init__(self, *args,
                 formatter: Optional[logging.Formatter] = DEFAULT_FORMATTER,
                 **kwargs):
        """Attach formatter to Handler instance."""
        super().__init__(*args, **kwargs)
        self.setFormatter(formatter)


class SingleMessageFormatHandlerMixin(FormattedHandlerMixin):
    """Mixin for logging.Handler derivatives. Assigns formatter at creation."""

    def __init__(self, *args,
                 formatter: Optional[logging.Formatter] = SINGLE_MESSAGE_FORMATTER,
                 **kwargs):
        """Attach formatter to Handler instance."""
        super().__init__(*args, formatter=formatter, **kwargs)


# handlers

class GtkBufferHandler(logging.Handler):
    def __init__(self, *args,
                 max_lines: Optional[int] = None,
                 **kwargs):
        """Abstract base class for any Gtk buffer handler.

        Limit maximum number of lines if max_lines set to int"""
        super().__init__(*args, **kwargs)
        self.max_lines = max_lines

    @property
    def max_lines(self, value: int):
        return self._max_lines

    @max_lines.setter
    def max_lines(self, value: Optional[int] = None):
        if value is None:
            self._max_lines = None # no limit
        elif isinstance(value, int) and value >= 0:
            self._max_lines = value
        else:
            raise ValueError(f"max_lines must be None or int, not {type(value)}.")

    def _insert_into_buffer(msg):
        ...

    def emit(self, record):
        """Emit log message to attached Gtk.TextBuffer."""
        try:
            msg = self.format(record)
            self._insert_into_buffer(msg)

        except Exception:
            self.handleError(record)


class GtkTextBufferHandler(GtkBufferHandler):
    def __init__(self, *args, text_buffer: Gtk.TextBuffer,
                 max_lines: Optional[int] = DEFAULT_TEXT_BUFFER_MAX_LINES,
                 **kwargs):
        """Tie a logging handler to a Gtk.TextBuffer.

        Limit maximum number of lines if max_lines set to int"""

        super().__init__(*args, max_lines=max_lines, **kwargs)
        self._buffer = text_buffer

        self._tag_bold = self._buffer.create_tag("bold", weight=Pango.Weight.BOLD)
        self._tag_italic = self._buffer.create_tag("italic", style=Pango.Style.ITALIC)


class GtkEntryBufferHandler(GtkBufferHandler):
    def __init__(self, *args, entry_buffer: Gtk.EntryBuffer,
                 max_lines: Optional[int] = DEFAULT_ENTRY_MAX_LINES, **kwargs):
        """Tie a logging handler to a Gtk.EntryBuffer.

        Limit maximum number of lines if max_lines set to int."""
        super().__init__(*args, max_lines=max_lines, **kwargs)
        self._buffer = entry_buffer


class GtkLabelHandler(GtkBufferHandler):
    def __init__(self, *args, label: Gtk.Label,
                 max_lines: Optional[int] = DEFAULT_LABEL_MAX_LINES, **kwargs):
        """Tie a logging handler to a Gtk.EntryBuffer.

        Limit maximum number of lines if max_lines set to int."""
        super().__init__(*args, max_lines=max_lines, **kwargs)
        self._buffer = label


class AppendingGtkTextBufferHandler(GtkTextBufferHandler):
    def _insert_into_buffer(self, msg):
        """Appends log message to attached Gtk.TextBuffer."""

        # remove all tags from buffer
        start_iter = self._buffer.get_start_iter()
        end_iter = self._buffer.get_end_iter()
        self._buffer.remove_tag(self._tag_bold, start_iter, end_iter)

        # insert new message at end
        self._buffer.insert_with_tags(end_iter, msg + "\n", self._tag_bold)

        # crop lines at end of buffer
        if (self._max_lines is not None
                and self._buffer.get_line_count() > self._max_lines):
            lines_to_remove = self._buffer.get_line_count() - self._max_lines
            start_iter = self._buffer.get_start_iter()
            end_iter = self._buffer.get_iter_at_line(lines_to_remove)
            self._buffer.delete(start_iter, end_iter)

        # place cursor at end
        self._buffer.place_cursor(self._buffer.get_end_iter())


class PrependingGtkTextBufferHandler(GtkTextBufferHandler):
    def _insert_into_buffer(self, msg):
        """Prepends log message to attached Gtk.TextBuffer."""

        # remove all tags from buffer
        start_iter = self._buffer.get_start_iter()
        end_iter = self._buffer.get_end_iter()
        self._buffer.remove_tag(self._tag_bold, start_iter, end_iter)

        # insert new message at beginning
        self._buffer.insert_with_tags(start_iter, msg + "\n", self._tag_bold)

        # crop lines at end of buffer
        if (self._max_lines is not None
                and self._buffer.get_line_count() > self._max_lines):
            lines_to_remove = self._buffer.get_line_count() - self._max_lines
            end_iter = self._buffer.get_end_iter()
            start_iter = self._buffer.get_iter_at_line(self._max_lines-lines_to_remove)
            self._buffer.delete(start_iter, end_iter)

        # place cursor at beginning
        self._buffer.place_cursor(self._buffer.get_start_iter())


class AppendingGtkEntryBufferHandler(GtkEntryBufferHandler):
    def _insert_into_buffer(self, msg):
        """Appends log message to attached Gtk.Entry."""

        # shorten message if too long
        if len(msg.splitlines()) > self._max_lines:
            new_text = "\n".join([
                *msg.splitlines()[:self._max_lines//2],
                '...'
                * msg.splitlines()[self._max_lines//2+1:]
            ])
        else: # crop older messages if too long
            original_text = self._buffer.get_text()
            new_text = original_text + "\n" + msg
            if (self._max_lines is not None
                    and len(new_text.splitlines()) > self._max_lines):
                lines_to_remove = len(new_text.splitlines()) - self._max_lines
                new_text = "\n".joint(new_text.splitlines()[lines_to_remove:])

        self._buffer.set_text(new_text)


class SingleMessageGtkEntryBufferHandler(GtkEntryBufferHandler):
    def _insert_into_buffer(self, msg):
        """Store log message in Gtk.Entry."""

        # shorten message if too long
        if len(msg.splitlines()) > self._max_lines:
            new_text = "\n".join([
                *msg.splitlines()[:self._max_lines//2],
                '...'
                * msg.splitlines()[self._max_lines//2+1:]
            ])
        else:
            new_text = msg
        self._buffer.set_text(new_text)


class SingleMessageGtkLabelHandler(GtkLabelHandler):
    """Show last line of log message on Gtk.Label."""
    def _insert_into_buffer(self, msg):
        """Store log message in Gtk.Label."""

        # shorten message if too long
        if len(msg.splitlines()) > self._max_lines:
            new_text = "\n".join(msg.splitlines()[-self._max_lines:])
        else:
            new_text = msg
        self._buffer.set_text(new_text)


class SingleMessageGtkInfoBarHandler(SingleMessageGtkLabelHandler):
    """Reveal attached Gtk.InfoBar when logging."""
    def __init__(self, *args, info_bar: Gtk.InfoBar, **kwargs):
        super().__init__(*args, **kwargs)
        self._info_bar = info_bar

    def emit(self, record):
        super().emit(record)
        self._info_bar.show()
        self._info_bar.set_revealed(True)


class FormattedAppendingGtkTextBufferHandler(FormattedHandlerMixin, AppendingGtkTextBufferHandler):
    pass


class FormattedPrependingGtkTextBufferHandler(FormattedHandlerMixin, PrependingGtkTextBufferHandler):
    pass


class FormattedSingleMessageGtkLabelHandler(SingleMessageFormatHandlerMixin, SingleMessageGtkLabelHandler):
    pass


class FormattedSingleMessageGtkInfoBarHandler(SingleMessageFormatHandlerMixin, SingleMessageGtkInfoBarHandler):
    pass



#
# Copyright 2020-2021 Lars Pastewka
#           2021 Johannes Hörmann
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import asyncio
import logging
import os

import yaml
from concurrent.futures import ProcessPoolExecutor

import dtoolcore
from dtoolcore.utils import generous_parse_uri
from dtool_info.utils import date_fmt, sizeof_fmt

from dtool_info.inventory import _dataset_info
from dtool_lookup_api.core.LookupClient import ConfigurationBasedLookupClient

from ..utils.multiprocessing import StatusReportingChildProcessBuilder, process_initializer
from ..utils.progressbar import ProgressBar


logger = logging.getLogger(__name__)


class CopyFuncWrapper:
    def __init__(self, copy_func):
        self._copy_func = copy_func

    def __call__(self, src_uri, dest_base_uri, status_report_callback):
        """Wraps a dtool copy_func into interface compatible with StatusReportingChildProcessBuilder"""
        self._copy_func(
            src_uri=src_uri,
            dest_base_uri=dest_base_uri,
            config_path=None,
            progressbar=status_report_callback,
        )


def _proto_dataset_info(dataset):
    """Return information about proto dataset as a dict."""
    # Analogous to dtool_info.inventory._dataset_info
    info = {}

    info['type'] = 'dtool-proto'

    info['uri'] = dataset.uri

    info['uuid'] = dataset.uuid

    info["size_int"] = None
    info["size_str"] = 'unknown'

    info['creator'] = dataset._admin_metadata['creator_username']
    info['name'] = dataset._admin_metadata['name']

    info["date"] = 'not yet frozen'

    info['readme_content'] = dataset.get_readme_content()

    return info


def _info(dataset):
    if isinstance(dataset, dtoolcore.DataSet):
        info = _dataset_info(dataset)
        info['type'] = 'dtool-dataset'
        info['is_frozen'] = True

        manifest = []
        for identifier in dataset._identifiers():
            manifest += [(identifier, dataset.item_properties(identifier))]
        info['manifest'] = manifest
    else:
        info = _proto_dataset_info(dataset)
        info['is_frozen'] = False

    p = generous_parse_uri(info['uri'])
    info['scheme'] = p.scheme
    info['base_uri'] = p.path if p.netloc is None else p.netloc

    return info


def _mangle_lookup_manifest(manifest_dict):
    """Convert dictionary returned from lookup server into a normalized manifest"""
    manifest = []
    for key, value in manifest_dict['items'].items():
        manifest += [(key, value)]
    return manifest


async def _lookup_info(lookup_dict):
    """Mangle return dict of lookup server into a proper dataset info"""

    info = {}

    uri = lookup_dict['uri']

    info['type'] = 'lookup'

    info['uri'] = uri
    p = generous_parse_uri(uri)
    info['scheme'] = p.scheme
    info['base_uri'] = p.path if p.netloc is None else p.netloc

    info['uuid'] = lookup_dict['uuid']

    # The server does not include these fields in response as of 0.17.2
    # They will be available in next release,
    # https://github.com/jic-dtool/dtool-lookup-server/pull/21
    try:
        info["size_int"] = lookup_dict['size_in_bytes']
        info["size_str"] = sizeof_fmt(lookup_dict['size_in_bytes'])
    except KeyError:
        info["size_int"] = None
        info["size_str"] = 'unknown'

    info['creator'] = lookup_dict['creator_username']
    info['name'] = lookup_dict['name']

    info['date'] = date_fmt(lookup_dict['frozen_at'])

    info['is_frozen'] = True

    return info


def _list_proto_datasets(base_uri):
    datasets = []
    for dataset in dtoolcore.iter_proto_datasets_in_base_uri(base_uri):
        datasets += [DatasetModel.from_dataset(dataset)]
    return datasets


def _list_datasets(base_uri):
    datasets = []
    for dataset in dtoolcore.iter_datasets_in_base_uri(base_uri):
        datasets += [DatasetModel.from_dataset(dataset)]
    return datasets


def _load_dataset(uri):
    logger.info(f'Loading dataset from URI: {uri}')

    # determine from admin metadata whether this is a protodataset
    admin_metadata = dtoolcore._admin_metadata_from_uri(uri, None)

    if admin_metadata['type'] == 'protodataset':
        dataset = dtoolcore.ProtoDataSet.from_uri(uri)
    else:
        dataset = dtoolcore.DataSet.from_uri(uri)

    return dataset


async def _copy_dataset(uri, target_base_uri, resume, auto_resume, progressbar=None):
    logger.info(f'Copying dataset from URI {uri} to {target_base_uri}...')

    dataset = _load_dataset(uri)

    dest_uri = dtoolcore._generate_uri(
        admin_metadata=dataset._admin_metadata,
        base_uri=target_base_uri
    )

    copy_func = dtoolcore.copy
    is_dataset = dtoolcore._is_dataset(dest_uri, config_path=None)
    if resume or (auto_resume and is_dataset):
        # copy resume
        copy_func = dtoolcore.copy_resume
    elif is_dataset:
        # don't resume
        raise FileExistsError("Dataset already exists: {}".format(dest_uri))
    else:
        # If the destination URI is a "file" dataset one needs to check if
        # the path already exists and exit gracefully if true.
        parsed_dataset_uri = dtoolcore.utils.generous_parse_uri(dest_uri)
        if parsed_dataset_uri.scheme == "file":
            if os.path.exists(parsed_dataset_uri.path):
                raise FileExistsError(
                    "Path already exists: {}".format(parsed_dataset_uri.path))

    num_items = len(list(dataset.identifiers))

    copy_func_wrapper = CopyFuncWrapper(copy_func)

    with ProgressBar(length=num_items,
                     label="Copying dataset",
                     pb=progressbar) as pb:
        non_blocking_copy_func = StatusReportingChildProcessBuilder(copy_func_wrapper, pb)
        dest_uri = await non_blocking_copy_func(uri, target_base_uri)

    logger.info(f'Dataset successfully copied from {uri} to {target_base_uri}.')

    return dest_uri


class DatasetModel:
    """
    Model for both frozen and proto datasets, either received from dtoolcore
    or the lookup server.
    """

    @staticmethod
    async def all(base_uri):
        """Return all datasets at base URI"""

        datasets = []

        loop = asyncio.get_running_loop()
        datasets = []
        with ProcessPoolExecutor(max_workers=2, initializer=process_initializer) as executor:
            datasets += await loop.run_in_executor(executor, _list_proto_datasets, base_uri)
            datasets += await loop.run_in_executor(executor, _list_datasets, base_uri)

        return datasets

    @classmethod
    def from_uri(cls, uri):
        return cls(uri=uri)

    @classmethod
    def from_dataset(cls, dataset):
        return cls(dataset_info=_info(dataset))

    @classmethod
    async def from_lookup(cls, lookup_dict):
        return cls(dataset_info=await _lookup_info(lookup_dict))

    def __init__(self, uri=None, dataset_info=None):
        if uri is not None:
            self.reload(uri)
        elif dataset_info is not None:
            self._dataset_info = dataset_info
        else:
            raise ValueError('Please provide either `uri` or `dateset_info`.')

    @classmethod
    async def search(cls, keyword):
        async with ConfigurationBasedLookupClient() as lookup:
            datasets = await lookup.search(keyword)

        return [await cls.from_lookup(lookup_dict) for lookup_dict in datasets]

    @classmethod
    async def query(cls, query_text):
        async with ConfigurationBasedLookupClient() as lookup:
            datasets = await lookup.by_query(query_text)

        return [await cls.from_lookup(lookup_dict) for lookup_dict in datasets]

    @classmethod
    async def query_all(cls):
        """Query all datasets from lookup server."""
        async with ConfigurationBasedLookupClient() as lookup:
            datasets = await lookup.all()

        return [await cls.from_lookup(lookup_dict) for lookup_dict in datasets]

    def __str__(self):
        return self._dataset_info['uri']

    def __getattr__(self, name):
        return self._dataset_info[name]

    def __setattr__(self, name, value):
        if name.startswith('_'):
            super().__setattr__(name, value)
        else:
            self._dataset_info[name] = value

    def __getstate__(self):
        return self._dataset_info

    def __setstate__(self, state):
        self._dataset_info = state

    def reload(self, uri=None):
        """Load the dataset from a URI.

        :param uri: URI to a dtoolcore.DataSet
        """
        self._dataset_info = _info(_load_dataset(uri))

    async def copy(self, target_base_uri, resume=False, auto_resume=True, progressbar=None):
        """Copy a dataset."""
        await _copy_dataset(self.uri, target_base_uri, resume, auto_resume, progressbar)

    def freeze(self):
        uri = str(self)
        _load_dataset(str(self)).freeze()
        logger.debug(f"Froze {uri}")
        # We need to reread dataset after freezing, since _data is currently
        # a dtoolcore.ProtoDataSet but should not become a dtoolcore.DataSet
        # Dataset and URI lost when freezing, preserve and reload here
        self.reload(uri=uri)
        uri = str(self)
        logger.debug(f"Reloaded {uri}")

    def put_readme(self, text):
        self.readme_content = text
        return _load_dataset(str(self)).put_readme(text)

    async def get_readme(self):
        if 'readme_content' in self._dataset_info:
            return self._dataset_info['readme_content']
        async with ConfigurationBasedLookupClient() as lookup:
            readme_dict = await lookup.readme(self.uri)
        self._dataset_info['readme_content'] = yaml.dump(readme_dict)
        return self._dataset_info['readme_content']

    async def get_manifest(self):
        # ATTENTION HERE: will try to get data from lookup server for proto datasets without following check
        if not self.is_frozen:
            return dict()
        if 'manifest' in self._dataset_info:
            return self._dataset_info['manifest']
        async with ConfigurationBasedLookupClient() as lookup:
            manifest_dict = await lookup.manifest(self.uri)
        self._dataset_info['manifest'] = _mangle_lookup_manifest(manifest_dict)
        return self._dataset_info['manifest']

    async def get_item(self, item_uuid):
        """Get item from dataset by item UUID"""
        if not self.is_frozen:
            raise ValueError("Cannot retrieve items by UUID from ProtoDatasets.")

        dataset = _load_dataset(str(self))
        if item_uuid in dataset.identifiers:
            cached_file = dataset.item_content_abspath(item_uuid)
            logger.debug(f"Retrieved cached item {cached_file}.")
            return cached_file
        else:
            raise ValueError(f"Item UUID '{item_uuid}' does not exist within dataset '{str(self)}'.")


#
# Copyright 2021 Lars Pastewka
#           2021 Johannes Hörmann
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

# the following import is necessary to patch flawed dtoolcore.utils function
import dtool_lookup_gui.utils.patch

import argparse
import asyncio
import logging
import sys

import gi
gi.require_version('Gtk', '3.0')
gi.require_version('GtkSource', '4')
from gi.repository import GLib, GObject, Gio, Gtk, GtkSource

import gbulb
gbulb.install(gtk=True)

from .views.main_window import MainWindow

# The following imports are need to register widget types with the GObject type system
import dtool_lookup_gui.widgets.base_uri_list_box
import dtool_lookup_gui.widgets.dataset_list_box
import dtool_lookup_gui.widgets.graph_widget
import dtool_lookup_gui.widgets.transfer_popover_menu
import dtool_lookup_gui.widgets.progress_chart
import dtool_lookup_gui.widgets.progress_popover_menu



logger = logging.getLogger(__name__)


# adapted from https://python-gtk-3-tutorial.readthedocs.io/en/latest/popover.html#menu-popover
class Application(Gtk.Application):
    def __init__(self, *args, loop=None, **kwargs):
        super().__init__(*args,
                         application_id='de.uni-freiburg.dtool-lookup-gui',
                         flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE, **kwargs)
        self.loop = loop
        self.args = None

    def do_activate(self):
        logger.debug("do_activate")

        # https://pyinstaller.readthedocs.io/en/latest/runtime-information.html
        if getattr(sys, 'frozen', False) and hasattr(sys, '_MEIPASS'):
            logger.debug('running in a PyInstaller bundle')
        else:
            logger.debug('running in a normal Python process')

        win = self.props.active_window
        if not win:
            # Windows are associated with the application
            # when the last one is closed the application shuts down
            # self.window = AppWindow(application=self, title="Main Window")
            logger.debug("Build GUI.")

            win = MainWindow(application=self)
            win.connect('destroy', lambda _: self.loop.stop())
            self.loop.call_soon(win.refresh)  # Populate widgets after event loop starts

        logger.debug("Present main window.")
        win.present()

    # adapted from http://fedorarules.blogspot.com/2013/09/how-to-handle-command-line-options-in.html
    # and https://python-gtk-3-tutorial.readthedocs.io/en/latest/application.html#example
    def do_command_line(self, args):
        """Handle command line options from within Gtk Application.

        Gtk.Application command line handler called if
        Gio.ApplicationFlags.HANDLES_COMMAND_LINE set.
        Must call self.activate() to get the application up and running."""

        Gtk.Application.do_command_line(self, args)  # call the default commandline handler

        # in order to have both:
        # * preformatted help text and ...
        # * automatic display of defaults
        class ArgumentDefaultsAndRawDescriptionHelpFormatter(
            argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter):
            pass

        parser = argparse.ArgumentParser(prog=self.get_application_id(),
                                         description=__doc__,
                                         formatter_class=ArgumentDefaultsAndRawDescriptionHelpFormatter)

        parser.add_argument('--verbose', '-v', action='count', dest='verbose',
                            default=0, help='Make terminal output more verbose')
        parser.add_argument('--debug', action='store_true',
                            help='Print debug info')
        parser.add_argument('--quiet','-q', action='store_true',
                            help='Print debug info')
        parser.add_argument('--log', required=False, nargs='?', dest="log",
                            default=None, const='out.log', metavar='LOG',
                            help='Write out.log, optionally specify log file name')

        # parse the command line stored in args, but skip the first element (the filename)
        self.args = parser.parse_args(args.get_arguments()[1:])

        loglevel = logging.WARNING
        logformat = "%(levelname)s: %(message)s"
        if self.args.quiet:
            loglevel = logging.ERROR
        if self.args.verbose > 0:
            loglevel = logging.INFO
        if self.args.debug or (self.args.verbose > 1):
            loglevel = logging.DEBUG

        if self.args.verbose > 2:
            logformat = (
                "[%(asctime)s - pid %(process)d - thread id %(thread)d - %(funcName)s - %(pathname)s:%(lineno)s]"
                " %(levelname)s: %(message)s"
            )

        # explicitly modify the root logger
        logging.basicConfig(level=loglevel, format=logformat)

        self.activate_action('set-loglevel', GLib.Variant.new_uint16(loglevel))
        if self.args.log:
            self.activate_action('set-logfile', GLib.Variant.new_string(self.args.log))

        logger.debug("Parsed CLI options {}".format(self.args))

        self.activate()
        return 0

    def do_startup(self):
        """Runs before anything else, create custom actions here."""
        logger.debug("do_startup")

        root_logger = logging.getLogger()

        # toggle-logging
        toggle_logging_variant = GLib.Variant.new_boolean(True)
        toggle_logging_action = Gio.SimpleAction.new_stateful(
            "toggle-logging", None, toggle_logging_variant
        )
        toggle_logging_action.connect("change-state", self.do_toggle_logging)
        self.add_action(toggle_logging_action)

        # set-loglevel
        loglevel_variant = GLib.Variant.new_uint16(root_logger.level)
        loglevel_action = Gio.SimpleAction.new_stateful(
            "set-loglevel", loglevel_variant.get_type(), loglevel_variant
        )
        loglevel_action.connect("change-state", self.do_set_loglevel)
        self.add_action(loglevel_action)

        # set-logfile
        logfile_variant = GLib.Variant.new_string('none')
        logfile_action = Gio.SimpleAction.new_stateful(
            "set-logfile", logfile_variant.get_type(), logfile_variant
        )
        logfile_action.connect("change-state", self.do_set_logfile)
        self.add_action(logfile_action)

        Gtk.Application.do_startup(self)

    # custom application-scoped actions
    def do_toggle_logging(self, action, value):
        action.set_state(value)
        if value.get_boolean():
            logger.debug("Return to default logging configuration.")
            logging.disable(logging.NOTSET)
            logger.debug("Returned to default logging configuration.")

        else:
            logger.debug("Disable all logging below WARNING.")
            logging.disable(logging.WARNING)
            logger.debug("Disabled all logging below WARNING.")

    def do_set_loglevel(self, action, value):
        loglevel = value.get_uint16()
        if action.get_state().get_uint16() == loglevel:
            logger.debug("Desired loglevel and current log level are equivalent.")
            return
        root_logger = logging.getLogger()
        root_logger.setLevel(loglevel)
        action.set_state(value)

    def do_set_logfile(self, action, value):
        logfile = value.get_string()
        if action.get_state().get_string() == logfile:
            logger.debug(f"Desired log file {logfile} and current log file are equivalent.")
            return
        fh = logging.FileHandler(logfile)
        root_logger = logging.getLogger()
        fh.setLevel(root_logger.level)
        fh.setFormatter(root_logger.handlers[0].formatter)
        root_logger.addHandler(fh)
        action.set_state(value)


def run_gui():
    GObject.type_register(GtkSource.View)

    loop = asyncio.get_event_loop()
    app = Application(loop=loop)
    logger.debug("do_startup")
    # see https://github.com/beeware/gbulb#gapplicationgtkapplication-event-loop
    loop.run_forever(application=app, argv=sys.argv)

